from typing import Callable
from numba import njit, prange
import numpy as np
from numpy import ndarray

from polymesh.utils import cells_coords

__cache = True


@njit(nogil=True, cache=__cache)
def element_coordinates(coords: ndarray, topo: np.ndarray):
    return cells_coords(coords, topo)


def element_centers(ecoords: np.ndarray):
    return np.mean(ecoords, axis=1)


@njit(nogil=True, parallel=True, cache=__cache)
def int_domain(
    points: np.ndarray,
    qpos: np.ndarray,
    qweight: np.ndarray,
    dshpfnc: Callable,
    weight: float = 1.0,
) -> ndarray:
    """
    Returns the measure (length, area or volume in 1d, 2d and 3d) of a
    single domain.
    """
    volume = 0.0
    points = np.transpose(points)
    nG = len(qweight)
    for iG in prange(nG):
        jac = points @ dshpfnc(qpos[iG])
        djac = np.linalg.det(jac)
        volume += qweight[iG] * djac
    return volume * weight


@njit(nogil=True, parallel=True, cache=__cache)
def int_domains(
    ecoords: np.ndarray, qpos: np.ndarray, qweight: np.ndarray, dshpfnc: Callable
):
    """
    Returns the measure (length, area or volume in 1d, 2d and 3d) of
    several domains.
    """
    nE = ecoords.shape[0]
    res = np.zeros(nE, dtype=ecoords.dtype)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        for i in prange(nE):
            jac = ecoords[i].T @ dshp
            djac = np.linalg.det(jac)
            res[i] += qweight[iG] * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix(
    C: np.ndarray,
    ecoords: np.ndarray,
    qpos: np.ndarray,
    qweight: np.ndarray,
    dshpfnc: Callable,
    sdmfnc: Callable,
    nDOFN=3,
):
    nTOTV = len(ecoords) * nDOFN
    res = np.zeros((nTOTV, nTOTV), dtype=C.dtype)
    points = np.transpose(ecoords)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        jac = points @ dshp
        djac = np.linalg.det(jac) * qweight[iG]
        B = sdmfnc(jac, dshp)
        res += B.T @ C @ B * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_bulk(
    C: np.ndarray,
    ecoords: np.ndarray,
    qpos: np.ndarray,
    qweight: np.ndarray,
    dshpfnc: Callable,
    sdmfnc: Callable,
    nDOFN=3,
):
    nE, nNODE, _ = ecoords.shape
    nTOTV = nNODE * nDOFN
    res = np.zeros((nE, nTOTV, nTOTV), dtype=C.dtype)
    nG = len(qweight)
    for iG in prange(nG):
        dshp = dshpfnc(qpos[iG])
        for iE in prange(nE):
            jac = np.transpose(ecoords[iE]) @ dshp
            djac = np.linalg.det(jac) * qweight[iG]
            B = sdmfnc(jac, dshp)
            res[iE] += B.T @ C[iE] @ B * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix_bulk2(B: ndarray, djac: ndarray, w: ndarray):
    nE, nG, nSTRE, nTOTV = B.shape
    res = np.zeros((nE, nSTRE, nTOTV), dtype=B.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += B[iE, iG] * djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def unit_strain_load_vector_bulk(D: ndarray, B: ndarray):
    nE, nSTRE, nTOTV = B.shape
    res = np.zeros((nE, nTOTV, nSTRE), dtype=D.dtype)
    for iE in prange(nE):
        res[iE] = B[iE].T @ D[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def strain_load_vector_bulk(BTD: ndarray, strains: ndarray):
    nE, nTOTV, nSTRE = BTD.shape
    nE, nRHS, nSTRE = strains.shape
    res = np.zeros((nE, nTOTV, nRHS), dtype=BTD.dtype)
    for iE in prange(nE):
        for iRHS in prange(nRHS):
            res[iE, :, iRHS] = BTD[iE] @ strains[iE, iRHS, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_bulk2(D: ndarray, B: ndarray, djac: ndarray, w: ndarray):
    nE, nG = djac.shape
    nTOTV = B.shape[-1]
    res = np.zeros((nE, nTOTV, nTOTV), dtype=D.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += B[iE, iG].T @ D[iE] @ B[iE, iG] * djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def mass_matrix_bulk(
    N: ndarray, dens: ndarray, weights: ndarray, djac: ndarray, w: ndarray
):
    """
    Evaluates the mass matrix of several elements from evaluations at the
    Gauss points of the elements. The densities are assumed to be constant
    over the elements, and be represented by the 1d float numpy array 'dens'.

    Parameters
    ----------
    N : 4d numpy float array
        Evaluations of the shape function matrix according to the type of
        the element for every cell and Gauss point.
    dens : 1d numpy float array
        1d float array of densities of several elements.
    weights : 1d numpy float array
        Array of weighting factors for the cells (eg. areas, volumes) with
        the same shape as 'dens'.
    w : 1d numpy float array
        Weights of the Gauss points
    djac : 2d numpy float array
        Jacobi determinants for every element and Gauss point.

    nE : number of elements
    nG : number of Gauss points
    nNE : number of nodes per element
    nDOF : number of dofs per node

    """
    nE, nG = djac.shape
    nTOTV = N.shape[-1]
    res = np.zeros((nE, nTOTV, nTOTV), dtype=N.dtype)
    for iG in range(nG):
        for iE in prange(nE):
            res[iE] += (
                dens[iE] * weights[iE] * N[iE, iG].T @ N[iE, iG] * djac[iE, iG] * w[iG]
            )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def compliance_bulk(K_bulk: np.ndarray, dofsol1d: np.ndarray, gnum: np.ndarray):
    """
    Returns the total compliance of the structure, that is
        u.T @ K @ u
    """
    nE, nTOTV, _ = K_bulk.shape
    res = np.zeros(nE, dtype=K_bulk.dtype)
    for iE in prange(nE):
        for i in prange(nTOTV):
            for j in prange(nTOTV):
                res[iE] += (
                    K_bulk[iE, i, j] * dofsol1d[gnum[iE, i]] * dofsol1d[gnum[iE, j]]
                )
    return res


@njit(nogil=True, cache=__cache)
def elastic_strain_energy_bulk(
    K_bulk: np.ndarray, dofsol1d: np.ndarray, gnum: np.ndarray
):
    """
    Returns the total elastic strain energy of the structure, that is
        0.5 * u.T @ K @ u
    """
    return compliance_bulk(K_bulk, dofsol1d, gnum) / 2


@njit(nogil=True, parallel=True, cache=__cache)
def approx_dof_solution(dofsol1d: np.ndarray, gnum: np.ndarray, shpmat: np.ndarray):
    nDOFN, nTOTV = shpmat.shape
    res = np.zeros((nDOFN), dtype=dofsol1d.dtype)
    for j in prange(nDOFN):
        for k in prange(nTOTV):
            res[j] += shpmat[j, k] * dofsol1d[gnum[k]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_dof_solution_bulk(
    dofsol1d: np.ndarray, gnum: np.ndarray, shpmat: np.ndarray
):
    nE = len(gnum)
    nDOFN, nTOTV = shpmat.shape
    res = np.zeros((nE, nDOFN), dtype=dofsol1d.dtype)
    for i in prange(nE):
        for j in prange(nDOFN):
            for k in prange(nTOTV):
                res[i, j] += shpmat[j, k] * dofsol1d[gnum[i, k]]
    return res


@njit(nogil=True, cache=__cache)
def element_dof_solution(dofsol1d: np.ndarray, gnum: np.ndarray):
    return dofsol1d[gnum]


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk(dofsol1d: ndarray, gnum: ndarray):
    nE, nEVAB = gnum.shape
    res = np.zeros((nE, nEVAB), dtype=dofsol1d.dtype)
    for i in prange(nE):
        res[i, :] = dofsol1d[gnum[i, :]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk_multi(dofsol: ndarray, gnum: ndarray):
    nRHS = dofsol.shape[-1]
    nE, nEVAB = gnum.shape
    res = np.zeros((nRHS, nE, nEVAB), dtype=dofsol.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            res[j, i, :] = dofsol[gnum[i, :], j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def avg_nodal_data_bulk(data: np.ndarray, topo: np.ndarray):
    nE, nNE, nD = data.shape
    nP = np.max(topo) + 1
    res = np.zeros((nP, nD), dtype=data.dtype)
    count = np.zeros((nP), dtype=topo.dtype)
    for iE in prange(nE):
        for jNE in prange(nNE):
            res[topo[iE, jNE]] += data[iE, jNE]
            count[topo[iE, jNE]] += 1
    for iP in prange(nP):
        res[iP] = res[iP] / count[iP]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_point_bulk(
    C: np.ndarray,
    dofsol1d: np.ndarray,
    ecoords: np.ndarray,
    gnum: np.ndarray,
    dshp: np.ndarray,
    sdmfnc: Callable,
):
    nE = len(gnum)  # number of elements
    nSTRE = C.shape[-1]  # number of generalized forces
    esol = element_dof_solution_bulk(dofsol1d, gnum)
    res = np.zeros((nE, nSTRE), dtype=C.dtype)
    for i in prange(nE):
        jac = ecoords[i].T @ dshp
        B = sdmfnc(jac, dshp)
        res[i] = C[i] @ B @ esol[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_cells_nodes(
    C: np.ndarray,
    dofsol1d: np.ndarray,
    ecoords: np.ndarray,
    gnum: np.ndarray,
    dshp: np.ndarray,
    sdmfnc: Callable,
):
    nE, nNODE, _ = ecoords.shape  # numbers of elements and element nodes
    nSTRE = C.shape[-1]  # number of generalized forces
    res = np.zeros((nE, nNODE, nSTRE), dtype=C.dtype)
    for i in prange(nNODE):
        res[:, i, :] = stresses_at_point_bulk(
            C, dofsol1d, ecoords, gnum, dshp[i], sdmfnc
        )
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def body_load_vector_bulk(
    values: ndarray, N: ndarray, djac: ndarray, w: ndarray
) -> ndarray:
    """
    Integrates the body loads to give an equivalent nodal load vector.
    This implementation is general in the context of Lagrange elements.

    Input values are assumed to be evaluated at multiple (nG) Gauss points of
    multiple (nE) cells.

    Parameters
    ----------
    values: numpy.ndarray
        A 3d float array of body loads of shape (nE, nRHS, nNE * nDOF)
        for several elements and load cases.
    N: numpy.ndarray
        A float array of shape (nG, nDOF, nDOF * nNODE), being the shape
        function matrix evaluated at a nG number of Gauss points of several
        cells.
    djac: numpy.ndarray
        A float array of shape (nE, nG), being jacobian determinants
        evaluated at the Gauss points of several cells.
    w: numpy.ndarray
        1d float array of weights for an nG number of Gauss points,
        with a shape of (nG,).

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nE, nNE * nDOF, nRHS).
    """
    nE, nRHS, nEVAB = values.shape
    nG = N.shape[0]
    res = np.zeros((nE, nEVAB, nRHS), dtype=values.dtype)
    for iG in range(nG):
        NTN = N[iG].T @ N[iG]
        for iRHS in prange(nRHS):
            for iE in prange(nE):
                res[iE, :, iRHS] += NTN @ values[iE, iRHS, :] * djac[iE, iG] * w[iG]
    return res



from typing import Iterable, Union

import numpy as np
from numpy import ndarray

from .ebc import _link_opposite_sides_sym, _link_opposite_sides
from .utils import (
    _strain_field_3d_bulk,
    _postproc_3d_gauss_stresses,
    _calc_avg_hooke_3d_to_shell,
)
from sigmaepsilon.fem import Structure
from sigmaepsilon.fem.ebc import NodeToNode


class RepresentativeVolumeElement(Structure):
    
    def __init__(self, *args, symmetric:bool=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.symmetric = symmetric
    
    def _periodic_essential_ebc(
        self, axis: Union[int, Iterable[int]] = 0,
    ) -> NodeToNode:
        if self.symmetric:
            return _link_opposite_sides_sym(self.mesh.points(), axis)
        else:
            return _link_opposite_sides(self.mesh, axis)
        
    def ABD(self) -> ndarray:
        """
        Returns the elasticity matrix for a Kirchhoff-Love plate.
        
        Returns
        -------
        numpy.ndarray
            The 6x6 elastic stiffness matrix.
        """
        return self.homogenize(target="KL")

    def ABDS(self) -> ndarray:
        """
        Returns the elasticity matrix for a Mindlin-Reissner plate.
        
        Returns
        -------
        numpy.ndarray
            The 8x8 elastic stiffness matrix.
        """
        return self.homogenize(target="MR")

    def homogenize(self, target: str = "MR") -> ndarray:
        """
        Returns the homogenized elasticity matrix for a Mindlin-Reissner or a
        Kirchhoff-Love plate.
        
        Parameters
        ----------
        target: str, Optional
            The target model. Accepted values are 'MR' for Mindlin-Reissner
            shells and 'KL' for Kirchhoff-Love shells.
        
        Returns
        -------
        numpy.ndarray
            The elastic stiffness matrix.
        """
        target = target.lower()
        if target in ["mr", "kl"]:
            return self._to_shell(target)
    
    def _to_shell(self, target: str = "MR") -> ndarray:
        mesh = self.mesh
        NDOFN = mesh.NDOFN

        if target.lower() == "mr":
            NSTRE = 8
        elif target.lower() == "kl":
            NSTRE = 6

        # mesh data
        points = mesh.points()
        [(xmin, xmax), (ymin, ymax), (_, zmax)] = points.bounds()
        area = (xmax - xmin) * (ymax - ymin)

        # nodal loads
        nodal_loads = np.zeros((len(points), NDOFN, NSTRE), dtype=float)
        mesh.pd.loads = nodal_loads

        # nodal supports against rigid body motion
        nodal_fixity = np.zeros((len(points), NDOFN), dtype=bool)
        i = points.index_of_closest([xmin, ymax, zmax])
        nodal_fixity[i, :3] = True
        i = points.index_of_closest([xmax, ymax, zmax])
        nodal_fixity[i, 2] = True
        i = points.index_of_closest([xmin, ymin, zmax])
        nodal_fixity[i, 2] = True
        i = points.index_of_closest([xmax, ymin, zmax])
        nodal_fixity[i, 1] = True
        mesh.pd.fixity = nodal_fixity

        # nodal supports to guarantee peridic displacement solution
        periodicity_constraints = self._periodic_essential_ebc(axis=[0, 1])

        # initial strain loads
        blocks = list(mesh.cellblocks(inclusive=True))
        for block in blocks:
            centers = block.cd.centers()
            strain_loads = np.zeros((centers.shape[0], 6, NSTRE))
            _strain_field_3d_bulk(centers, out=strain_loads, NSTRE=NSTRE)
            if block.cd.NDIM == 3:
                block.cd.strain_loads = strain_loads
            elif block.cd.NDIM == 2:
                raise NotImplementedError
            else:
                raise NotImplementedError

        # solve BVP
        self.constraints.append(periodicity_constraints)
        self.linear_static_analysis()

        # postproc
        hooke = np.zeros((NSTRE, NSTRE), dtype=float)
        hooke_avg = np.zeros_like(hooke)
        for block in blocks:
            block.cd.strain_loads = None
            hooke_block = block.cd.elastic_material_stiffness_matrix()
            gp, gw = block.cd.quadrature["full"]
            if block.cd.NDIM == 3:
                cell_forces = block.cd.internal_forces(points=gp, flatten=False)
                # cell_forces = block.internal_forces(points=gp, flatten=False, target='global')
                ec = block.cd.coords()
                dshp = block.cd.shape_function_derivatives(gp)
                jac = block.cd.jacobian_matrix(dshp=dshp, ecoords=ec)
                djac = block.cd.jacobian(jac=jac)
                ec = block.cd.coords(points=gp)
                _postproc_3d_gauss_stresses(cell_forces, ec, gw, djac, hooke)
                _calc_avg_hooke_3d_to_shell(hooke_block, ec, gw, djac, hooke_avg)
            elif block.cd.NDIM == 2:
                raise NotImplementedError
            else:
                raise NotImplementedError

        self.constraints.pop(-1)

        hooke /= area
        hooke_avg /= area
        result = hooke_avg - hooke
        if NSTRE == 8:
            result[6:8, 6:8] = hooke[6:8, 6:8] * 5 / 6
        result = (result + result.T) / 2
        return result



from polymesh.cells import T6 as Triangle
from polymesh.utils.cells.gauss import (
    Gauss_Legendre_Tri_1,
    Gauss_Legendre_Tri_3a,
    Gauss_Legendre_Tri_3b,
)

from ..material.membrane import Membrane
from ..material.mindlinplate import MindlinPlate
from ..material.mindlinshell import MindlinShell

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class LST_M(ABC, Membrane, Triangle, FiniteElement):
    """
    Linear strain triangle or LST. Also called quadratic triangle and Veubeke
    triangle. Developed by Fraeijs de Veubeke in 1962-63, published in 1965 [1].

    References
    ----------
    .. [1] B. M. Fraeijs de Veubeke, Displacement and equilibrium models, in Stress
       Analysis, ed. by O. C. Zienkiewicz and G. Hollister, Wiley, London, 1965, 145-197.
       Reprinted in Int. J. Numer. Meth. Engrg., 52, 287-342, 2001.
    """

    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Tri_3a(),
        "full2": Gauss_Legendre_Tri_3b(),
        "selective": {(0, 1): "full", (2,): "reduced"},
        "reduced": Gauss_Legendre_Tri_1(),
    }


class LST_P_MR(ABC, MindlinPlate, Triangle, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Tri_3a(),
        "full2": Gauss_Legendre_Tri_3b(),
        "selective": {(0, 1, 2): "full", (3, 4): "reduced"},
        "reduced": Gauss_Legendre_Tri_1(),
    }


class LST_S_MR(ABC, MindlinShell, Triangle, FiniteElement):
    qrule = "selective"
    quadrature = {
        "full": Gauss_Legendre_Tri_3a(),
        "full2": Gauss_Legendre_Tri_3b(),
        "selective": {(0, 1, 3, 4, 5): "full", (2, 6, 7): "reduced"},
        "reduced": Gauss_Legendre_Tri_1(),
    }



import json
import numpy as np
from numpy import ndarray, average as avg
from typing import Union, Iterable, Any
from numbers import Number

from dewloosh.core.tools import float_to_str_sig
from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools.dtk import parsedicts_addr
from neumann.function import Function

from .problem import NavierProblem
from .preproc import rhs_rect_const, rhs_conc_1d, rhs_conc_2d, rhs_line_const
from .utils import points_to_rectangle_region, sin1d, cos1d


class NavierLoadError(Exception):
    """
    Exception raised for invalid load inputs.
    """

    def __init__(self, message=None):
        if message is None:
            message = (
                "Invalid input for loads. "
                "It must be a dictionary, a NumPy array, or "
                "an instance of `sigmaepsilon.fourier.LoadGroup`."
            )
        super().__init__(message)


class LoadGroup(LinkedDeepDict):
    """
    A class to handle load groups for Navier's semi-analytic solution of
    rectangular plates and beams with specific boundary conditions.

    This class is also the base class of all other load types.

    See Also
    --------
    :class:`LinkedDeepDict`

    Examples
    --------

    >>> from sigmaepsilon.fourier import LoadGroup, PointLoad
    >>> loads = LoadGroup(
    >>>     group1 = LoadGroup(
    >>>         case1 = PointLoad(x=L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=L/3, v=[0.0, 1.0]),
    >>>     ),
    >>>     group2 = LoadGroup(
    >>>         case1 = PointLoad(x=2*L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=2*L/3, v=[0.0, 1.0]),
    >>>     ),
    >>> )

    Since the LoadGroup is a subclass of LinkedDeepDict,
    a case is accessible as

    >>> loads['group1', 'case1']

    If you want to protect the object from the accidental
    creation of nested subdirectories, you can lock the layout
    by typing

    >>> loads.lock()
    """

    _typestr_ = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__problem = None

    @property
    def problem(self) -> NavierProblem:
        """
        Returns the attached problem.
        """
        if self.is_root():
            return self.__problem
        return self.root().problem

    @problem.setter
    def problem(self, value: NavierProblem):
        """
        Sets the attached problem.

        Parameters
        ----------
        value : NavierProblem
            The problem the loads are defined for.
        """
        assert self.is_root(), "The problem can only be set on the top-level object."
        self.__problem = value

    def blocks(
        self,
        *args,
        inclusive: bool = False,
        blocktype: Any = None,
        deep: bool = True,
        **kwargs,
    ) -> Iterable["LoadGroup"]:
        """
        Returns a generator object that yields all the subgroups.

        Parameters
        ----------
        inclusive : bool, Optional
            If True, returns the object the call is made upon.
            Default is False.
        blocktype : Any, Optional
            The type of the load groups to return. Default is None, that
            returns all types.
        deep : bool, Optional
            If True, all deep groups are returned separately. Default is True.

        Yields
        ------
        LoadGroup
        """
        dtype = LoadGroup if blocktype is None else blocktype
        return self.containers(self, inclusive=inclusive, dtype=dtype, deep=deep)

    def cases(self, *args, inclusive: bool = True, **kwargs) -> Iterable["LoadGroup"]:
        """
        Returns a generator that yields the load cases in the group.

        Parameters
        ----------
        inclusive : bool, Optional
            If True, returns the object the call is made upon.
            Default is True.
        blocktype : Any, Optional
            The type of the load groups to return. Default is None, that
            returns all types.
        deep : bool, Optional
            If True, all deep groups are returned separately. Default is True.

        Yields
        ------
        LoadGroup
        """
        return filter(
            lambda i: i.__class__._typestr_ is not None,
            self.blocks(*args, inclusive=inclusive, **kwargs),
        )

    @staticmethod
    def _string_to_dtype_(string: str = None):
        if string == "group":
            return LoadGroup
        elif string == "rectangle":
            return RectangleLoad
        elif string == "point":
            return PointLoad
        elif string == "line":
            return LineLoad
        else:
            raise NotImplementedError()

    def dump(self, path: str, *, mode: str = "w", indent: int = 4):
        """
        Dumps the content of the object to a file.

        Parameters
        ----------
        path : str
            The path of the file on your filesystem.
        mode : str, Optional
            https://www.programiz.com/python-programming/file-operation
        indent : int, Optional
            Governs the level to which members will be pretty-printed.
            Default is 4.
        """
        with open(path, mode) as f:
            json.dump(self.to_dict(), f, indent=indent)

    @classmethod
    def from_json(cls, path: str = None) -> "LoadGroup":
        """
        Loads a loadgroup from a JSON file.

        Parameters
        ----------
        path : str
            The path to a file on your filesystem.

        Returns
        -------
        LoadGroup
        """
        if path is not None:
            with open(path, "r") as f:
                d = json.load(f)
            return cls.from_dict(d)

    def _encode_(self, *_, **__) -> dict:
        """
        Returns the group as a dictionary. Overwrite this in child
        implementations.
        """
        res = {}
        cls = type(self)
        res = {
            "type": cls._typestr_,
            "key": self.key,
        }
        return res

    @classmethod
    def _decode_(cls, d: dict = None, *_, **kwargs):
        """
        Returns a LoadGroup object from a dictionary.
        Overwrite this in child implementations.

        Parameters
        ----------
        d : dict, Optional
            A dictionary.
        **kwargs : dict, Optional
            Keyword arguments defining a load group.
        """
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            "key": d.pop("key", None),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def to_dict(self) -> dict:
        """
        Returns the group as a dictionary.
        """
        res = self._encode_()
        for key, value in self.items():
            if isinstance(value, LoadGroup):
                res[key] = value.to_dict()
            else:
                res[key] = value
        return res

    @staticmethod
    def from_dict(d: dict = None, **kwargs) -> "LoadGroup":
        """
        Reads a LoadGroup from a dictionary. The keys in the dictionaries
        must match the parameters of the corresponding load types and a type
        indicator.
        """
        if "type" in d:
            cls = LoadGroup._string_to_dtype_(d["type"])
        else:
            cls = LoadGroup
        res = cls(**d)
        for addr, value in parsedicts_addr(d, inclusive=True):
            if len(addr) == 0:
                continue
            if "type" in value:
                cls = LoadGroup._string_to_dtype_(value["type"])
            else:
                cls = LoadGroup
            value["key"] = addr[-1]
            res[addr] = cls(**value)
        return res

    def __repr__(self):
        return "LoadGroup(%s)" % (dict.__repr__(self))


class RectangleLoad(LoadGroup):
    """
    A class to handle rectangular loads.

    Parameters
    ----------
    v: Iterable
       Load intensities for each dof in the order :math:`fz, mx, my`.
    x: Iterable, Optional
       The coordinates of the lower-left and upper-right points of the region
       where the load is applied. Default is None.
    """

    _typestr_ = "rectangle"

    def __init__(self, *args, x: Iterable = None, v: Iterable = None, **kwargs):
        super().__init__(*args, x=x, v=v, **kwargs)

    @classmethod
    def _decode_(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            "key": d.pop("key", None),
            "x": np.array(d.pop("x")),
            "v": np.array(d.pop("v")),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            "type": cls._typestr_,
            "key": self.key,
            "x": float_to_str_sig(self["x"], sig=6),
            "v": float_to_str_sig(self["v"], sig=6),
        }
        return res

    def region(self) -> Iterable:
        """
        Returns the region as a list of 4 values x0, y0, w, and h, where x0 and y0 are
        the coordinates of the bottom-left corner, w and h are the width and height
        of the region.
        """
        assert "x" in self, "There are no points defined."
        return points_to_rectangle_region(self["x"])

    def rhs(self, *, problem: NavierProblem = None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.

        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        p = problem if problem is not None else self.problem
        x = np.array(self["x"], dtype=float)
        v = np.array(self["v"], dtype=float)
        return rhs_rect_const(p.size, p.shape, x, v)

    def __repr__(self):
        return "RectangleLoad(%s)" % (dict.__repr__(self))


class LineLoad(LoadGroup):
    """
    A class to handle loads over lines for both beam and plate problems.

    Parameters
    ----------
    x : Iterable
        The point of application as an 1d iterable for a beam, a 2d iterable
        for a plate. In the latter case, the first row is the first point, the
        second row is the second point.
    v : Iterable
        Load intensities for each dof. The order of the dofs for a beam
        is [F, M], for a plate it is [F, Mx, My].
    """

    _typestr_ = "line"

    def __init__(self, *args, x: Iterable = None, v: Iterable = None, **kwargs):
        super().__init__(*args, x=x, v=v, **kwargs)

    @classmethod
    def _decode_(cls, d: dict = None, *_, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            "key": d.pop("key", None),
            "x": np.array(d.pop("x")),
            "v": np.array(d.pop("v")),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            "type": cls._typestr_,
            "key": self.key,
            "x": float_to_str_sig(self["x"], sig=6),
            "v": float_to_str_sig(self["v"], sig=6),
        }
        return res

    def rhs(self, *, problem: NavierProblem = None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.
        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        p = problem if problem is not None else self.problem
        x = np.array(self["x"], dtype=float)
        v = self["v"]
        if len(x.shape) == 1:
            assert len(v) == 2, f"Invalid shape {v.shape} for load intensities."
            if isinstance(v[0], Number) and isinstance(v[1], Number):
                v = np.array(v, dtype=float)
                return rhs_line_const(p.length, p.N, v, x)
            else:
                rhs = np.zeros((1, p.N, 2), dtype=x.dtype)
                if isinstance(v[0], str):
                    f = Function(v[0], variables=["x"], dim=1)
                    L = p.length
                    points = np.linspace(x[0], x[1], 1000)
                    d = x[1] - x[0]
                    rhs[0, :, 0] = list(
                        map(
                            lambda i: (2 / L)
                            * d
                            * avg(sin1d(points, i, L) * f([points])),
                            np.arange(1, p.N + 1),
                        )
                    )
                elif isinstance(v[0], Number):
                    _v = np.array([v[0], 0], dtype=float)
                    rhs[0, :, 0] = rhs_line_const(p.length, p.N, _v, x)[0, :, 0]
                if isinstance(v[1], str):
                    f = Function(v[1], variables=["x"], dim=1)
                    L = p.length
                    points = np.linspace(x[0], x[1], 1000)
                    d = x[1] - x[0]
                    rhs[0, :, 1] = list(
                        map(
                            lambda i: (2 / L)
                            * d
                            * avg(cos1d(points, i, L) * f([points])),
                            np.arange(1, p.N + 1),
                        )
                    )
                elif isinstance(v[1], Number):
                    _v = np.array([0, v[1]], dtype=float)
                    rhs[0, :, 1] = rhs_line_const(p.length, p.N, _v, x)[0, :, 1]
                return rhs

    def __repr__(self):
        return "LineLoad(%s)" % (dict.__repr__(self))


class PointLoad(LoadGroup):
    """
    A class to handle concentrated loads.

    Parameters
    ----------
    x : Union[float, Iterable]
        The point of application. A scalar for a beam, an iterable of
        length 2 for a plate.
    v : Iterable
        Load values for each dof. The order of the dofs for a beam
        is [F, M], for a plate it is [F, Mx, My].
    """

    _typestr_ = "point"

    def __init__(
        self, *args, x: Union[float, Iterable] = None, v: Iterable = None, **kwargs
    ):
        super().__init__(*args, x=x, v=v, **kwargs)

    @classmethod
    def _decode_(cls, d: dict = None, *_, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            "key": d.pop("key", None),
            "x": np.array(d.pop("x")),
            "v": np.array(d.pop("v")),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            "type": cls._typestr_,
            "key": self.key,
            "x": float_to_str_sig(self["x"], sig=6),
            "v": float_to_str_sig(self["v"], sig=6),
        }
        return res

    def rhs(self, *, problem: NavierProblem = None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.
        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        p = problem if problem is not None else self.problem
        x = self["x"]
        v = np.array(self["v"])
        if hasattr(p, "size"):
            return rhs_conc_2d(p.size, p.shape, v, x)
        else:
            return rhs_conc_1d(p.length, p.N, v, x)

    def __repr__(self):
        return "PointLoad(%s)" % (dict.__repr__(self))



from typing import Iterable

import numpy as np
from numpy import ndarray

from neumann import atleastnd
from polymesh.abcdata import ABCMeta_MeshData
from polymesh.cell import PolyCell

from ..dofmap import DOF
from ...utils.fem.fem import (
    expand_shape_function_matrix_bulk,
    element_dofmap_bulk,
)


class FemMixin:
    dofs = ()
    dofmap = ()

    # optional
    qrule: str = None
    quadrature = None

    # advanced settings
    compatible = True

    # must be reimplemented
    NNODE: int = None  # number of nodes, normally inherited

    # from the mesh object
    # it affects sizing of all kinds of intermediate variables
    # during calculations (eg. element transformation matrix)
    NDOFN: int = None  # numper of dofs per node, normally

    # inherited form the model object
    NDIM: int = None  # number of geometrical dimensions,
    # normally inherited from the mesh object

    # inherited form the model object
    NSTRE: int = None  # number of internal force components,

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def geometry(self):
        pd, cd = self.pdb, self.db
        return self.__class__.Geometry(pointdata=pd, celldata=cd)

    @property
    def pdb(self):
        return self.pointdata

    @property
    def db(self):
        return self._wrapped

    def topology(self) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    @classmethod
    def lcoords(cls, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    @classmethod
    def lcenter(cls, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def local_coordinates(self) -> ndarray:
        # implemented at PolyCell
        raise NotImplementedError

    def points_of_cells(self) -> ndarray:
        # implemented at PolyCell
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def jacobian_matrix(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def jacobian(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at material
    def strain_displacement_matrix(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    def shape_function_values(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    # !TODO : can be reimplemented if the element is not IsoP
    def shape_function_matrix(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at geometry
    def shape_function_derivatives(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at material
    def elastic_material_stiffness_matrix(self) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at material
    def stresses_at_centers(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    # !TODO : this should be implemented at material
    def stresses_at_cells_nodes(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def direction_cosine_matrix(self) -> ndarray:
        return None

    def integrate_body_loads(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def masses(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def areas(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def lengths(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def volumes(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def densities(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def weights(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError

    def dof_solution(self, *, cells: Iterable, flatten: bool) -> ndarray:
        raise NotImplementedError

    def kinetic_strains(self, *, cells: Iterable, points: Iterable) -> ndarray:
        raise NotImplementedError
    
    def centers(self, *args, **kwargs) -> ndarray:
        raise NotImplementedError


class FemMaterial(FemMixin):
    """
    Base class for FEM material classes.
    """


class MetaFiniteElement(ABCMeta_MeshData):
    """
    Python metaclass for safe inheritance. Throws a TypeError
    if a method tries to shadow a definition in any of the base
    classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)
        for base in bases:
            if issubclass(base, PolyCell):
                cls.Geometry = base
            if issubclass(base, FemMaterial):
                if len(base.dofs) > 0:
                    cls.Material = base
                    cls.dofs = base.dofs
                    cls.NDOFN = len(cls.dofs)
                    cls.dofmap = np.array(DOF.dofmap(cls.dofs), dtype=int)
        return cls


class ABCFiniteElement(FemMixin, metaclass=MetaFiniteElement):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """

    __slots__ = ()
    dofs = ()
    dofmap = ()

    def shape_function_matrix(self, *args, expand: bool = False, **kwargs):
        N = super().shape_function_matrix(*args, **kwargs)
        if expand:
            constant_metric = len(N.shape) == 3
            N = atleastnd(N, 4, front=True)
            nDOFN = self.container.NDOFN
            dofmap = self.__class__.dofmap
            if len(dofmap) < nDOFN:
                # the model has more dofs than the element
                nE, nP, nX = N.shape[:3]
                nTOTV = nDOFN * self.NNODE
                # nE, nP, nX, nDOF * nNE
                N_ = np.zeros((nE, nP, nX, nTOTV), dtype=float)
                dofmap = element_dofmap_bulk(dofmap, nDOFN, self.NNODE)
                N = expand_shape_function_matrix_bulk(N, N_, dofmap)
            return N[0] if constant_metric else N
        else:
            return N



try:
    import pypardiso

    __haspardiso__ = True
except Exception:
    __haspardiso__ = False


try:
    import sympy

    __hassympy__ = True
except Exception:
    __hassympy__ = False


try:
    import pylatex

    __haspylatex__ = True
except Exception:
    __haspylatex__ = False



import numpy as np
from numpy import ndarray

from neumann.linalg import Tensor4x3
from neumann.linalg.top import tr_3333_jit


class ComplianceTensor(Tensor4x3):
    def __init__(self, *args, tensorial: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        if self._array and not tensorial:
            collapsed = self.collapsed
            self.collapse()
            self._array = _compliance_to_tensorial(self._array)
            if not collapsed:
                self.expand()

    def transform_components(self, dcm: np.ndarray) -> ndarray:
        """
        Returns the components of the transformed numerical tensor,
        based on the provided direction cosine matrix.
        """
        if self.collapsed:
            self.expand()
            array = tr_3333_jit(self._array, dcm)
            self.collapse()
        else:
            array = tr_3333_jit(self._array, dcm)
        return array

    def engineering_components(self) -> ndarray:
        """
        Returns the engineering components.
        """
        collapsed = self.collapsed
        self.collapse()
        result = _compliance_from_tensorial(self._array)
        if not collapsed:
            self.expand()
        return result


def _compliance_to_tensorial(S: ndarray) -> ndarray:
    R_inv = np.diag([1, 1, 1, 1 / 2, 1 / 2, 1 / 2], dtype=float)
    return R_inv @ S


def _compliance_from_tensorial(S: ndarray) -> ndarray:
    R = np.diag([1, 1, 1, 2, 2, 2], dtype=float)
    return R @ S



from polymesh.cells import QuadraticLine as Line
from polymesh.utils.cells.gauss import Gauss_Legendre_Line_Grid

from .bernoulli import BernoulliBase as Bernoulli
from .gen.b3 import (
    shape_function_values_bulk,
    shape_function_derivatives_bulk,
    shape_function_derivatives_multi_L,
)

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


__all__ = ["Bernoulli3"]


class Bernoulli3(ABC, Bernoulli, Line, FiniteElement):
    """
    Finite element class to handle 3-noded Bernoulli beams.
    """

    qrule = "full"
    quadrature = {
        "full": Gauss_Legendre_Line_Grid(6),
        "mass": Gauss_Legendre_Line_Grid(8),
    }
    shpfnc = shape_function_values_bulk
    dshpfnc = shape_function_derivatives_bulk
    dshpfnc_geom = shape_function_derivatives_multi_L



import numpy as np
from numpy import ndarray
from copy import deepcopy
from typing import NamedTuple, Iterable

from neumann.linalg.sparse import csr_matrix as csr
from polymesh.utils import cells_around
from sigmaepsilon.fem.structure import Structure
from sigmaepsilon.fem.femsolver import StaticSolver

from .filter import sensitivity_filter, sensitivity_filter_csr
from .utils import (
    get_filter_factors,
    get_filter_factors_csr,
    weighted_stiffness_flat as weighted_stiffness,
    element_stiffness_ranges,
)
from ...utils.fem.postproc import element_compliances_flat as element_compliances


class OptRes(NamedTuple):
    x: Iterable  # the design variables
    obj: float  # the actual value of the objective function
    vol: float  # the actual volume
    pen: float  # the actual value of the penalty
    n: int  # the number of iterations completed


def maximize_stiffness(
    structure: Structure,
    *_,
    miniter: int = 50,
    maxiter: int = 100,
    p_start: float = 1.0,
    p_stop: float = 3.0,
    p_inc: float = 0.2,
    p_step: int = 5,
    q: float = 0.5,
    vfrac: float = 0.6,
    dtol: float = 0.1,
    r_max: float = None,
    penalty: float = None,
    nostop: bool = True,
    neighbours: Iterable = None,
    guess: Iterable = None,
    i_start: int = 0,
    **__
) -> OptRes:
    """
    Performs topology optimization using an Optimality Criteria Method to
    maximize the stiffness of a structure, given a design space and a certain 
    amount of material to distribute.
    
    .. math::
        :nowrap:
        
        \\begin{equation}
            \\begin{array}{rrclcl}
            \\displaystyle \\min_{\\boldsymbol{\\rho}} & \\mathbf{u(\\boldsymbol{\\rho})}^T \\mathbf{f} \\\\
            \\textrm{s. t.} \\\\
            & \\mathbf{K}(\\boldsymbol{\\rho}) \\mathbf{u(\\boldsymbol{\\rho})} & = & \\mathbf{f} \\\\
            & V(\\boldsymbol{\\rho}) - \\eta V_0 & \\leq & 0 & & \\\\
            & \\rho_i \\in \\{0, 1\\} & & & & \\forall i \\in N
            \\end{array}
        \\end{equation}
    
    Parameters
    ----------
    structure : Structure
        An instance of sigmaepsilon.fem.Structure.
    miniter : int, Optional
        The minimum number of iterations to perform. Default is 50.
    maxiter : int, Optional
        The maximum number of iterations to perform. Default is 100.
    p_start : float, Optional
        Initial value of the penalty on intermediate densities. Default is 1.
    p_stop : float, Optional
        Final value of the penalty on intermediate densities. Default is 3.
    p_inc : float, Optional
        Increment of the penalty on intermediate densities. Default is 0.2
    p_step : int, Optional
        The number of interations it takes to increment the penalty on intermediate
        density values. Default is 5.
    q : float, Optional
        Smoothing factor. Defaul is 0.5. 
    vfrac : float, Optional
        The fraction of available volume and the volume of the virgin structure.
        Default is 0.6. 
    dtol : float, Optional
        This controls the maximum change in the value of a design variable. 
        Default is 0.1. 
    r_max : float, Optional
        Radius for filtering. Default is None.
    neighbours : float, Optional
        The neighbours of the cells for filtering. Default is None.
    guess : numpy.ndarray, Optional
        A guess on the solution. This parameter can be used to contiue
        a workflow. Default is None.  
    i_start : int, Optional
        Starting index for iterations. This parameter can be used to contiue
        a workflow. Default is 0.
    summary : bool, Optional
        If True, a short summary about execution time and the number of iterations
        is available after execution as `structure.summary['topopt']`. Default is False.
    nostop : bool, Optional
        If True, iterations neglect all stopping criteria, those govern by 
        mniniter and maxiter included.
        Default is False.
    
    Yields
    ------
    OptRes
        The results of the actual iteration.
        
    Notes
    -----
    - The function returns a generator expression.
    - This function can be used for both size and topology optimization,
      depending on the inputs.
    """
    do_filter = r_max is not None

    if structure._static_solver_ is None:
        structure.linear_static_analysis()
    femsolver: StaticSolver = structure._static_solver_.core
    assert femsolver.regular
    krows, kcols = femsolver.krows, femsolver.kcols
    kshape = femsolver.kshape
    kranges = element_stiffness_ranges(kshape)
    K_virgin = np.copy(femsolver.K_bulk.flatten())
    mesh = structure.mesh
    vols = mesh.volumes()
    centers = mesh.centers()

    # initial solution to set up parameters
    dens = np.zeros_like(vols)
    dens_tmp = np.zeros_like(dens)
    dens_tmp_ = np.zeros_like(dens)
    dCdx = np.zeros_like(dens)
    comps = np.zeros_like(dens)

    def get_dof_solution():
        U = femsolver.u
        dU = len(U.shape)
        if dU == 2:
            # multiple load cases
            return U[:, 0]
        return U

    def compliance(update_stiffness: bool = False) -> ndarray:
        if update_stiffness:
            femsolver.update_stiffness(weighted_stiffness(K_virgin, dens, kranges))
        femsolver._proc_()
        U = get_dof_solution()
        comps[:] = element_compliances(K_virgin, U, krows, kcols, kranges)
        np.clip(comps, 1e-7, None, out=comps)
        return np.sum(comps)

    # ------------------ INITIAL SOLUTION ---------------
    comp = compliance()
    vol = np.sum(vols)
    vol_start = vol
    vol_min = vfrac * vol_start

    # initialite filter
    if do_filter:
        if neighbours is None:
            neighbours = cells_around(centers, r_max, frmt="dict")
        if isinstance(neighbours, csr):
            factors = get_filter_factors_csr(centers, neighbours, r_max)
            fltr = sensitivity_filter_csr
        else:
            factors = get_filter_factors(centers, neighbours, r_max)
            fltr = sensitivity_filter

    # initialize penalty parameters
    if penalty is not None:
        p_start = penalty
        p_stop = penalty + 1
        p_step = maxiter + 1

    # ------------- INITIAL FEASIBLE SOLUTION ------------
    if guess is None:
        dens[:] = vfrac
    else:
        dens = deepcopy(guess)
    vol = np.sum(dens * vols)
    comp = compliance(update_stiffness=True)
    cIter = i_start
    p = p_start
    yield OptRes(dens, comp, vol, p, cIter)

    # ------------------- ITERATION -------------------
    terminate = False
    while not terminate:
        if (p < p_stop) and (np.mod(cIter, p_step) == 0):
            p += p_inc

        # estimate lagrangian
        lagr = p * comp / vol

        # set up boundaries of change
        _dens = dens * (1 - dtol)
        np.clip(_dens, 1e-5, 1.0, out=_dens)
        dens_ = dens * (1 + dtol)
        np.clip(dens_, 1e-5, 1.0, out=dens_)

        # sensitivity [*(-1)]
        dCdx[:] = p * comps * dens ** (p - 1)

        # sensitivity filter
        if do_filter:
            dCdx[:] = fltr(dCdx, dens, neighbours, factors)

        # calculate new densities and lagrangian
        dens_tmp_[:] = dens * (dCdx / vols) ** q
        dens_tmp[:] = dens_tmp_
        _lagr = 0
        lagr_ = 2 * lagr
        _maxtries = 200
        _ntries = 0
        while (lagr_ - _lagr) > 1e-3:
            if _ntries == _maxtries:
                raise RuntimeError("Couldn't find multiplier :(.")
            _lagr_ = (_lagr + lagr_) / 2
            dens_tmp[:] = dens_tmp_ / (_lagr_**q)
            np.clip(dens_tmp, _dens, dens_, out=dens_tmp)
            vol_tmp = np.sum(dens_tmp * vols)
            if vol_tmp < vol_min:
                lagr_ = _lagr_
            else:
                _lagr = _lagr_
            _ntries += 1
        lagr = lagr_
        dens[:] = dens_tmp

        # resolve equilibrium equations and calculate compliance
        comp = compliance(update_stiffness=True)
        vol = np.sum(dens * vols)
        cIter += 1
        res = OptRes(dens, comp, vol, p, cIter)
        yield res

        if nostop:
            terminate = False
        else:
            if cIter < miniter:
                terminate = False
            elif cIter >= maxiter:
                terminate = True
            else:
                terminate = p >= p_stop

    femsolver.update_stiffness(K_virgin)
    return OptRes(dens, comp, vol, p, cIter)



from typing import Iterable, Union

import numpy as np
from numpy import ndarray

from polymesh.utils.space import index_of_closest_point
from polymesh.space import PointCloud
from polymesh.utils import locate_tri_2d
from polymesh.cells import T3 as Triangle

from sigmaepsilon.fem import FemMesh
from sigmaepsilon.fem.ebc import NodeToNode
from sigmaepsilon.fem.ebc import BodyToBody


def _link_opposite_sides_sym(
    points: PointCloud, axis: Union[int, Iterable[int]] = 0
) -> NodeToNode:
    coords = points.show()
    bounds = points.bounds()
    links = []
    if isinstance(axis, int):
        axis = [axis]
    assert isinstance(
        axis, Iterable
    ), "'axis' must be an integer or an 1d iterable of integers."
    for axid in axis:
        bmin, bmax = bounds[axid]
        mask_source = coords[:, axid] < (bmin + 1e-12)
        i_source = np.where(mask_source)[0]
        source = coords[i_source]
        mask_target = coords[:, axid] > (bmax - 1e-12)
        i_target = np.where(mask_target)[0]
        target = coords[i_target]
        i = index_of_closest_point(source, target)
        id_source = points[i_source].id[i]
        id_target = points[i_target].id
        links.append(np.stack([id_source, id_target], axis=1))
    return NodeToNode(np.vstack(links))


def _link_opposite_sides(
    mesh: FemMesh, axis: Union[int, Iterable[int]] = 0
) -> ndarray:
    if isinstance(axis, int):
        axis = [axis]
    assert isinstance(
        axis, Iterable
    ), "'axis' must be an integer or an 1d iterable of integers."
        
    surface = mesh.surface()
    points = surface.points()
    triangles = surface.topology()
    centers = surface.centers()
    coords = points.show()
    bounds = points.bounds()
    
    def _link_points_to_cells_(source_indices, target_indices):
        source_coords = coords[source_indices]
        target_triangles = triangles[target_indices]
        
        i_source, i_target, locations_target = \
            locate_tri_2d(source_coords[:, mask], coords[:, mask], target_triangles)
        source_indices = source_indices[i_source]   
        target_triangles = target_triangles[i_target]
        
        shp_target = Triangle.shape_function_values(locations_target)
        
        nE, nNE = target_triangles.shape
        factors = np.zeros((nE, nNE + 1), dtype=float)
        indices = np.zeros((nE, nNE + 1), dtype=int)
        factors[:, -1] = -1
        factors[:, :nNE] = shp_target
        indices[:, -1] = source_indices
        indices[:, :nNE] = target_triangles
        
        return factors, indices
    
    factors, indices = [], []
    
    for axid in axis:
        bmin, bmax = bounds[axid]
        
        mask = np.ones(3, dtype=bool)
        mask[axid] = False
        
        mask_source = coords[:, axid] < (bmin + 1e-12)
        source_indices = np.where(mask_source)[0]
        mask_target = centers[:, axid] > (bmax - 1e-12)
        target_indices = np.where(mask_target)[0]
        f, i = _link_points_to_cells_(source_indices, target_indices)
        factors.append(f)
        indices.append(i)
        
        mask_source = coords[:, axid] > (bmax - 1e-12)
        source_indices = np.where(mask_source)[0]
        mask_target = centers[:, axid] < (bmin + 1e-12)
        target_indices = np.where(mask_target)[0]
        f, i = _link_points_to_cells_(source_indices, target_indices)
        factors.append(f)
        indices.append(i)
        
    return BodyToBody(factors=np.vstack(factors), indices=np.vstack(indices))



from typing import Iterable, Union, Tuple, Callable
from abc import abstractmethod
import numpy as np
from numpy import ndarray, concatenate as conc
from scipy.sparse import coo_matrix as coo

from neumann.linalg import ReferenceFrame
from neumann import atleast1d, atleast2d, repeat_diagonal_2d

from polymesh.space import PointCloud
from polymesh.cell import PolyCell3d
from polymesh.utils.space import index_of_closest_point

from .mesh import FemMesh
from .dofmap import DOF
from .constants import DEFAULT_DIRICHLET_PENALTY
from ..utils.fem.ebc import (
    link_points_to_points, 
    link_points_to_body,
    _body_to_body_stiffness_data_,
)


__all__ = ["NodalSupport", "NodeToNode", "FaceToFace", "BodyToBody"]


class EssentialBoundaryCondition:
    """
    Base class for Dirichlet boundary conditions accounted for
    using Courant-type penalization.
    """
            
    @abstractmethod
    def assemble(self, mesh: FemMesh) -> Tuple[coo, ndarray]:
        ...


class BodyToBody(EssentialBoundaryCondition):
    """
    Constrains the dofs of two touching bodies by gluing them together.
    
    Parameters
    ----------
    source: PolyCell
        The source body.
    target: PolyCell
        The target body.
    dofs: Iterable, Optinal
        An iterable of the constrained global degrees of freedom. It not specified, 
        all degrees of freedom are constrained.
    penalty: float, Optional
        The penalty value.
    lazy: bool, Optional
        Default is True.
    tol: float, Optional
        Floating point tolerance for detecting point in polygons. Default is 1e-12.
    k: int, Optional
        THe number of neighbours.
        
    Notes
    -----
    The two bodies must have a common surface.
    """
    def __init__(
        self,
        source: PolyCell3d=None,
        target: PolyCell3d=None,
        dofs: Iterable = None,
        penalty: float = DEFAULT_DIRICHLET_PENALTY,
        lazy: bool = True,
        tol: float = 1e-12,
        k: int = 4,
        touching: bool=False,
        factors: ndarray=None,
        indices: ndarray=None
    ):  
        if source and target:
            assert source.NDIM == 3, "Source must be a 3 dimensional body!"
            assert target.NDIM == 3, "Source must be a 3 dimensional body!"
            assert source.container.source() is target.container.source(), (
                "The source and the target must belong to the same pointcloud!"
            )
            assert source.container.source() is source.container.root(), (
                "The mesh must be brought to a standard form!"
            )
            assert isinstance(target.__class__.monomsfnc, Callable), (
                "The class is not equipped with the tools for this operation."
            )
        self.source = source
        self.target = target
        self.penalty = penalty
        self.lazy = lazy
        self.k = k
        self.tol = tol
        self.touching = touching
        self.factors=factors
        self.indices=indices
        
        if isinstance(dofs, Iterable):
            self.dofmap = DOF.dofmap(dofs)
        else:
            self.dofmap = None
    
    def assemble(self, mesh: FemMesh) -> Tuple[coo, ndarray]: 
        if (self.factors is None) or (self.indices is None):
            S: PolyCell3d = self.source
            T: PolyCell3d = self.target
            
            assert S.container.root() == mesh, \
                "The input mesh must be the root of both the source and the target."
            assert T.container.root() == mesh, \
                "The input mesh must be the root of both the source and the target."
            
            if self.touching:
                coords, topo_source_surface = S.extract_surface(detach=False)
                source_indices = np.unique(topo_source_surface)
            else:
                coords = S.source_coords()
                source_indices = S.unique_indices()
            
            source_coords = coords[source_indices]
            factors, indices = link_points_to_body(
                PointCloud(source_coords, inds=source_indices),
                T, self.lazy, self.tol, self.k
            )
        else:
            factors, indices = self.factors, self.indices
            
        nDOF = mesh.NDOFN
        nN = len(mesh.pointdata)
        N = nDOF * nN
        if self.dofmap is None:
            dmap = np.arange(nDOF)
        else:
            dmap = self.dofmap
        dmap = np.array(dmap, dtype=int)
        
        factors, indices = \
            _body_to_body_stiffness_data_(factors, indices, dmap, nDOF)
        fdata = factors.flatten()
        frows = np.repeat(np.arange(factors.shape[0]), factors.shape[1])
        fcols = indices.flatten()
        factors = coo((fdata, (frows, fcols)), shape=(factors.shape[0], N))
                
        Kp = self.penalty * (factors.T @ factors)
        fp = np.zeros(N, dtype=float)
        Kp.eliminate_zeros()
        return Kp, fp
            
    
class NodeToNode(EssentialBoundaryCondition):
    """
    Constrains relative motion of nodes.

    Parameters
    ----------
    imap : Union[dict, ndarray, list]
        An iterable describing pairs of nodes.
    penalty : float, Optional
        Penalty value for Courant-type penalization.
        Default is `~sigmaepsilon.fem.constants.DEFAULT_DIRICHLET_PENALTY`.
    dofs: Iterable, Optinal
        An iterable of the constrained degrees of freedom. It not specified, all
        degrees of freedom are constrained.
    source: PointCloud, Optional
        The source pointcloud. Only if 'imap' is not provided.
    target: PointCloud, Optional
        The target pointcloud. Only if 'imap' is not provided. 

    Example
    -------
    The following lines tie together all DOFs of nodes 1 with node 2 and node 3 with 4.
    The large penalty value means that the tied nodes should have the same displacements.

    >>> from sigmaepsilon.fem import NodeToNode
    >>> n2n = NodeToNode([[1, 2], [3, 4]], penalty=1e12)

    To tie only DOFs 'UX' and 'ROTZ':

    >>> n2n = NodeToNode([[1, 2], [3, 4]], dofs=['UX', 'ROTZ'], penalty=1e12)
    """

    def __init__(
        self,
        imap: Union[dict, ndarray, list] = None,
        *,
        source: PointCloud = None,
        target: PointCloud = None,
        dofs: Iterable = None,
        penalty: float = DEFAULT_DIRICHLET_PENALTY,
    ):
        if imap is None:
            if isinstance(source, PointCloud) and isinstance(target, PointCloud):
                imap = link_points_to_points(source, target)
        self.imap = imap
        self.penalty = penalty
        if isinstance(dofs, Iterable):
            self.dofmap = DOF.dofmap(dofs)
        else:
            self.dofmap = None

    def assemble(self, mesh: FemMesh) -> Tuple[coo, ndarray]:
        """
        Returns the penalty stiffness matrix and the penalty load matrix.
        """
        imap = None
        if isinstance(self.imap, dict):
            imap = np.stack([list(self.imap.keys()), list(self.imap.values())], axis=1)
        else:
            imap = np.array(self.imap).astype(int)
        nI = len(imap)
        
        nDOF = mesh.NDOFN
        nN = len(mesh.pointdata)
        N = nDOF * nN
        
        if self.dofmap is None:
            dmap = np.arange(nDOF)
        else:
            dmap = self.dofmap
        dmap = np.array(dmap, dtype=int)
        
        nF = nI * len(dmap)
        fdata = np.tile([1, -1], nF)
        frows = np.repeat(np.arange(nF), 2)
        i_source = conc([dmap + (i_ * nDOF) for i_ in imap[:, 0]])
        i_target = conc([dmap + (i_ * nDOF) for i_ in imap[:, 1]])
        fcols = np.stack([i_source, i_target], axis=1).flatten()
        factors = coo((fdata, (frows, fcols)), shape=(nF, N))
        
        Kp = self.penalty * (factors.T @ factors)
        fp = np.zeros(N, dtype=float)
        Kp.eliminate_zeros()
        return Kp, fp


class NodalSupport(EssentialBoundaryCondition):
    """
    A class to handle nodal supports.

    Parameters
    ----------
    data : dict, Optional
        Dictionary to define the prescribed values. Valid keys are
        'UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ'. Default is None.
    x : Iterable, Optional
        An iterable expressing the location of one or more points.
        Use it if you don't know the index of the point of application
        in advance. Default is None.
    i : Iterable[int] or int, Optional
        The index of one or more points of in a pointcloud.
        Default is None.
    frame : numpy.ndarray or ReferenceFrame, Optional
        Use it do define inclined supports. If it is a NumPy array,
        it must be a 3x3 matrix representing a set of orthonormal base
        vectors. A None value means that the prescribed values are understood
        in the frame of the mesh the constraints are applied to.
        Default is None.
    penalty : float, Optional
        Penalty value for Courant-type penalization.
        Default is `~sigmaepsilon.fem.constants.DEFAULT_DIRICHLET_PENALTY`.
    **kwargs : dict, Optional
        Keyword arguments used for elementwise definition of 'data'.

    Examples
    --------
    To define a support at one node

    >>> from sigmaepsilon.fem import NodalSupport
    >>> support = NodalSupport(x=[10, 0, 0], UZ=1.)

    To define an inclined support with nonzero prescribed displacements:

    >>> from polymesh.space import StandardFrame
    >>> GlobalFrame = StandardFrame(dim=3)
    >>> frame = GlobalFrame.rotate_new('Body', [0, angle, 0], 'XYZ')
    >>> support = NodalSupport(x=[10, 0, 0], frame=frame, UX=1., UZ=1.)

    To apply the same constrain at multiple locations:

    >>> support = NodalSupport(x=[[0, 0, 0], [L, 0, 0]], UX = 0., UY=0.)

    To define a support with a custom penalty:

    >>> support = NodalSupport(x=[[0, 0, 0], [L, 0, 0]],
    >>>                        UX = 0., UY=0., penalty=1e12)

    The prescribed values can also be provided with a dictionary as the first
    positional argument:

    >>> support = NodalSupport({'UX' : 0., 'UY' : 0.}, x=[[0, 0, 0], [L, 0, 0]],
    >>>                         penalty=1e12)
    """

    def __init__(
        self,
        data: dict = None,
        *,
        x: Iterable = None,
        i: Union[int, Iterable[int]] = None,
        frame: Union[ndarray, ReferenceFrame] = None,
        penalty: float = DEFAULT_DIRICHLET_PENALTY,
        **kwargs,
    ):
        self.x = x if x is None else np.array(x, dtype=float)
        self.i = i if i is None else np.array(i, dtype=int)
        self.penalty = penalty
        self.data = data
        self.frame = frame
        if data is None:
            self.data = kwargs
        self.dofmap = DOF.dofmap(self.data.keys())
        if not isinstance(self.i, (int, Iterable)):
            assert self.x is not None, "Index or position must be defined!"
        if self.frame is not None:
            if not isinstance(self.frame, ReferenceFrame):
                raise TypeError(
                    "Invalid frame type. Read the docs of the NodalSupport class."
                )
        else:
            self.frame = ReferenceFrame(dim=3)

    def assemble(self, mesh: FemMesh) -> Tuple[coo, ndarray]:
        """
        Returns the penalty stiffness matrix and the penalty load matrix.
        """
        i = None
        if self.i is None:
            coords = mesh.coords()
            x = atleast2d(self.x, front=True)
            i = index_of_closest_point(coords, x)
        else:
            i = atleast1d(self.i)
        nI = len(i)
        nDOF = mesh.NDOFN
        nN = len(mesh.pointdata)
        N = nDOF * nN
        c, r = divmod(nDOF, 3)
        assert r == 0, (
            "The number of deegrees of freedom per node must be a multiple of 3."
        )
        dcm = self.frame.dcm(source=mesh.frame)
        nodal_dcm = repeat_diagonal_2d(dcm, c)[self.dofmap]
        factors = repeat_diagonal_2d(nodal_dcm, nI)
        dofs = np.arange(nDOF)
        inds = conc([dofs + (i_ * nDOF) for i_ in i])
        kdata = (factors.T @ factors).flatten()
        krows = np.repeat(inds, len(inds))
        kcols = np.tile(inds, len(inds))
        Kp = self.penalty * coo((kdata, (krows, kcols)), shape=(N, N))
        fp = np.zeros(N, dtype=float)
        fdata = factors.T @ np.tile(list(self.data.values()), nI)
        fp[inds] = self.penalty * fdata
        Kp.eliminate_zeros()
        return Kp, fp



import numpy as np

from neumann import atleastnd, ascont
from polymesh.section import LineSection

from ....utils.material.bernoulli import calc_beam_stresses_2d, calc_beam_stresses_4d
from ....core.material import MaterialLike


class BeamSection(LineSection, MaterialLike):
    """
    Wraps an instance of `sectionproperties.analysis.section.Section` and
    adds a little here and there to make some of the functionality more
    accessible.

    Parameters
    ----------
    *args : tuple, Optional
        The first parameter can be a string referring to a section type, or an
        instance of :class:`sectionproperties.analysis.section.Section`. In the
        former case, parameters of the section can be provided as keyword arguments,
        which are then forwarded to :func:`get_section`, see its documentation
        for further details.
    **kwargs : dict, Optional
        The parameters of a section as keyword arguments, only if the first
        positional argument is a string (see above).
    mesh_params : dict, Optional
        A dictionary controlling the density of the mesh of the section.
        Default is None.
    material : :class:`sectionproperties.pre.pre.Material`, Optional
        The material of the section. If not specified, a default material is
        used. Default is None.
    wrap : :class:`sectionproperties.analysis.section.Section`
        A section object to be wrapped. It can also be provided as the
        first positional argument.
    Notes
    -----
    The implementation here only covers homogeneous cross sections. If you want
    to define an inhomogeneous section, it must be explicity given either as
    the first position argument, or with the keyword argument `wrap`.

    Examples
    --------
    >>> from sigmaepsilon import BeamSection
    >>> section = BeamSection(get_section('CHS', d=1.0, t=0.1, n=64))

    or simply provide the shape as the first argument and everything
    else with keyword arguments:

    >>> section = BeamSection('CHS', d=1.0, t=0.1, n=64)

    Plot a section with Matplotlib using 6-noded triangles:

    >>> import matplotlib.pyplot as plt
    >>> from dewloosh.mpl import triplot
    >>> section = BeamSection('CHS', d=1.0, t=0.3, n=32,
    >>>                       mesh_params=dict(n_max=20))
    >>> triobj = section.trimesh(T6=True).to_triobj()
    >>> fig, ax = plt.subplots(figsize=(4, 2))
    >>> triplot(triobj, fig=fig, ax=ax, lw=0.1)
    """

    def calculate_section_properties(self, separate: bool = False) -> dict:
        """
        Retruns a dictionary containing the properties of the section.
        """
        if not separate:
            # 0-xx 1-yy 2-zz 3-yz 4-xz 5-xy
            strs = np.zeros((6, 6, len(self.mesh["vertices"])))
            """strs = {'xx': {}, 'yy': {}, 'zz': {}, 'xy': {}, 'xz': {}, 'yz': {}}"""
        else:
            raise NotImplementedError

        self.calculate_geometric_properties()
        section_properties = self.geometric_properties

        self.calculate_warping_properties()
        _strs = np.eye(6)
        for i in range(6):
            N, Vy, Vx, Mzz, Mxx, Myy = _strs[i]
            stress_post = self.calculate_stress(
                N=N, Vy=Vy, Vx=Vx, Mzz=Mzz, Mxx=Mxx, Myy=Myy
            )
            stresses = stress_post.get_stress()
            if not separate:
                # this assumes that the section is homogeneous, hence the 0 index
                strs[0, i, :] = stresses[0]["sig_zz"]  # xx
                strs[5, i, :] = stresses[0]["sig_zx"]  # xy
                strs[4, i, :] = stresses[0]["sig_zy"]  # xz
            else:
                raise NotImplementedError

        section_properties["stress-factors"] = strs
        self.props = section_properties
        return section_properties

    def elastic_stiffness_matrix(self, E: float = None, nu: float = None) -> np.ndarray:
        """
        Returns the elastic stiffness matrix of the section.

        Parameters
        ----------
        E : float, Optional
            Young's modulus. If not provided, the value is obtained
            from the associated material of the section, if there is one.
            Default is None.
        nu : float, Optional
            Poisson's ratio. If not provided, the value is obtained
            from the associated material of the section, if there is one.
            Default is None.

        Returns
        -------
        numpy.ndarray
            The 4x4 elastic ABDS matrix of the section.
        """
        if E is None:
            E = self.geometry.material.elastic_modulus
            nu = self.geometry.material.poissons_ratio
        G = E / (2 * (1 + nu))
        return np.array(
            [
                [E * self.A, 0, 0, 0],
                [0, G * self.Ix, 0, 0],
                [0, 0, E * self.Iy, 0],
                [0, 0, 0, E * self.Iz],
            ]
        )

    def calculate_stresses(self, forces: np.ndarray) -> np.ndarray:
        """
        Returns stresses for all points of the mesh. The internal forces
        are expected in the order Fx, Fy, Fz, Mx, My, Mz, where x, y and z
        denote the local axes of the section, the stress components are
        returned in the order s11, s22, s33, s12, s13, s23.

        Parameters
        ----------
        forces : numpy.ndarray
            It can be either

                A 1d float array representing all internal forces of a beam.

                A 2d float array. In this case it is assumed, that the provided
                data has the shape (nX, nSTRE), where 'nX' is the number of records
                in the data.

                A 3d float array. In this case it is assumed, that the provided
                data has the shape (nE, nP, nSTRE), representing
                all internal forces (nSTRE) for multiple elements(nE) and
                evaluation points (nP).

                A 4d float array of shape (nE, nP, nSTRE, nRHS), representing
                all internal forces (nSTRE) for multiple elements(nE), load cases(nRHS)
                and evaluation points (nP).

        Returns
        -------
        numpy.ndarray
            A numpy float array that can be

                A 2d array of shape (nPS, nS),

                A 5d array of shape (nE, nP, nPS, nS, nRHS),

            where

                nPS : the number of points of the mesh of the section

                nS : the number of stress components

                nE : the number of finite elements

                nP : the number of evaluation points for an element

                nRHS : the number of right hand sides (possibly load cases)

        Notes
        -----
        This only covers homogeneous cross sections.
        """
        assert isinstance(forces, np.ndarray)
        factors = self.props["stress-factors"]
        nD = len(forces.shape)
        if nD == 1:
            f_ = ascont(atleastnd(forces, n=2, front=True))
            return calc_beam_stresses_2d(f_, factors)[0, ...]
        elif nD == 2:
            return calc_beam_stresses_2d(forces, factors)
        elif nD == 3:
            f_ = ascont(atleastnd(forces, n=4, back=True))
            return calc_beam_stresses_4d(f_, factors)[..., 0]
        elif nD == 4:
            return calc_beam_stresses_4d(forces, factors)
        else:
            raise RuntimeError("Unrecognizable shape of internal forces.")



import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


def _strain_field_3d_bulk(
    centers: ndarray, *, out: ndarray = None, NSTRE: int = 8
) -> ndarray:
    if out is None:
        out = np.zeros((centers.shape[0], 6, NSTRE), dtype=float)
    else:
        NSTRE = out.shape[-1]
    # case 1 - exx_0
    out[:, 0, 0] = 1.0  # exx
    # case 2 - eyy_0
    out[:, 1, 1] = 1.0  # eyy
    # case 3 - exy_0
    out[:, 5, 2] = 1.0  # exy
    # case 4 - kxx
    out[:, 0, 3] = centers[:, 2]  # exx
    # case 5 - kyy
    out[:, 1, 4] = centers[:, 2]  # eyy
    # case 6 - kxy
    out[:, 5, 5] = centers[:, 2]  # exy
    if NSTRE == 8:
        # case 7 - exz_0
        out[:, 4, 6] = 1.0  # exz
        # case 8 - eyz_0
        out[:, 3, 7] = 1.0  # eyz
    return out


def _expand_strain_arrays_3d(loads: ndarray) -> ndarray:
    nE, _, nS2d = loads.shape
    res = np.zeros((nE, 3, 3, nS2d), dtype=float)
    res[:, 0, 0, :] = loads[:, 0, :]
    res[:, 1, 1, :] = loads[:, 1, :]
    res[:, 2, 2, :] = loads[:, 2, :]
    res[:, 1, 2, :] = loads[:, 3, :] / 2
    res[:, 0, 2, :] = loads[:, 4, :] / 2
    res[:, 0, 1, :] = loads[:, 5, :] / 2
    res[:, 2, 1, :] = loads[:, 3, :] / 2
    res[:, 2, 0, :] = loads[:, 4, :] / 2
    res[:, 1, 0, :] = loads[:, 5, :] / 2
    return res


def _collapse_strain_arrays_3d(loads: ndarray) -> ndarray:
    nE, _, _, nS2d = loads.shape
    res = np.zeros((nE, 6, nS2d), dtype=float)
    res[:, 0, :] = loads[:, 0, 0, :]
    res[:, 1, :] = loads[:, 1, 1, :]
    res[:, 2, :] = loads[:, 2, 2, :]
    res[:, 3, :] = loads[:, 1, 2, :] * 2
    res[:, 4, :] = loads[:, 0, 2, :] * 2
    res[:, 5, :] = loads[:, 0, 1, :] * 2
    return res


def _strain_arrays_3d_to_Mindlin(loads: ndarray) -> ndarray:
    nE, _, nS2d = loads.shape
    res = np.zeros((nE, 8, nS2d), dtype=float)
    res[:, 0, :] = loads[:, 0, :]
    res[:, 1, :] = loads[:, 1, :]
    res[:, 2, :] = loads[:, 5, :]
    res[:, 6, :] = loads[:, 4, :]
    res[:, 7, :] = loads[:, 3, :]
    return res


def _strain_arrays_3d_to_Kirchhoff(loads: ndarray) -> ndarray:
    nE, _, nS2d = loads.shape
    res = np.zeros((nE, 6, nS2d), dtype=float)
    res[:, 0, :] = loads[:, 0, :]
    res[:, 1, :] = loads[:, 1, :]
    res[:, 2, :] = loads[:, 5, :]
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def _postproc_shell_gauss_dynams(
    cell_dynams: ndarray,  # cell dynams
    weights: ndarray,  # gauss weights
    djac: ndarray,  # jacobian determinants
    out: ndarray,  # ABD or ABDS matrix of shape (6, 6) or (8, 8)
) -> ndarray:
    nE = cell_dynams.shape[0]
    nP = weights.shape[0]
    NSTRE = out.shape[0]
    mindlin = NSTRE == 8
    for iE in range(nE):
        for iP in range(nP):
            w = weights[iP]
            dj = djac[iE, iP]
            for j in prange(NSTRE):
                nxx, nyy, nxy, mxx, myy, mxy, vxz, vyz = cell_dynams[iE, iP, :, j]
                out[0, j] += nxx * w * dj
                out[1, j] += nyy * w * dj
                out[2, j] += nxy * w * dj
                out[3, j] += mxx * w * dj
                out[4, j] += myy * w * dj
                out[5, j] += mxy * w * dj
                if mindlin:
                    out[6, j] += vxz * w * dj
                    out[7, j] += vyz * w * dj
    return out


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _postproc_3d_gauss_stresses(
    cell_stresses: ndarray,  # cell stresses
    cell_gauss_coords: ndarray,  # cell gauss coordinates
    weights: ndarray,  # gauss weights
    djac: ndarray,  # jacobian determinants
    out: ndarray,  # ABD or ABDS matrix of shape (6, 6) or (8, 8)
) -> ndarray:
    nE, nP = cell_gauss_coords.shape[:2]
    NSTRE = out.shape[0]
    mindlin = NSTRE == 8
    for iE in range(nE):
        for iP in range(nP):
            w = weights[iP]
            z = cell_gauss_coords[iE, iP, 2]
            dj = djac[iE, iP]
            for j in range(NSTRE):
                sxx, syy, _, syz, sxz, sxy = cell_stresses[iE, iP, :, j]
                out[0, j] += sxx * w * dj
                out[1, j] += syy * w * dj
                out[2, j] += sxy * w * dj
                out[3, j] += sxx * z * w * dj
                out[4, j] += syy * z * w * dj
                out[5, j] += sxy * z * w * dj
                if mindlin:
                    out[6, j] += sxz * w * dj
                    out[7, j] += syz * w * dj
    return out


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _calc_avg_hooke_shell(
    hooke: ndarray,  # cell material stiffness matrices
    areas: ndarray,  # gauss weights
    out: ndarray,  # ABD or ABDS matrix of shape (6, 6) or (8, 8)
) -> ndarray:
    nE = areas.shape()[0]
    for iE in range(nE):
        out += areas[iE] * hooke[iE]
    return out


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _calc_avg_hooke_3d_to_shell(
    hooke: ndarray,  # cell material stiffness matrices
    cell_gauss_coords: ndarray,  # cell gauss coordinates
    weights: ndarray,  # gauss weights
    djac: ndarray,  # jacobian determinants
    out: ndarray,  # ABD or ABDS matrix of shape (6, 6) or (8, 8)
) -> ndarray:
    C_126 = np.zeros((3, 3), dtype=hooke.dtype)
    C_45 = np.zeros((2, 2), dtype=hooke.dtype)
    nE, nP = cell_gauss_coords.shape[:2]
    NSTRE = out.shape[0]
    mindlin = NSTRE == 8
    for iE in range(nE):
        C11 = hooke[iE, 0, 0]
        C12 = hooke[iE, 0, 1]
        C22 = hooke[iE, 1, 1]
        C66 = hooke[iE, 5, 5]
        C55 = hooke[iE, 4, 4]
        C44 = hooke[iE, 3, 3]
        C_126[0, 0] = C11
        C_126[0, 1] = C12
        C_126[1, 0] = C12
        C_126[1, 1] = C22
        C_126[2, 2] = C66
        if mindlin:
            C_45[0, 0] = C55
            C_45[1, 1] = C44
        for iP in range(nP):
            w = weights[iP]
            z = cell_gauss_coords[iE, iP, 2]
            dj = djac[iE, iP]
            out[:3, :3] += C_126 * w * dj
            out[:3, 3:6] += C_126 * z * w * dj
            out[3:6, :3] += C_126 * z * w * dj
            out[3:6, 3:6] += C_126 * z**2 * w * dj
            if mindlin:
                out[6:, 6:] += C_45 * w * dj
    return out



from typing import Union
import numpy as np
import sympy as sy
from numpy import ndarray
from numpy.linalg import inv

from ..hooke.utils import _get_elastic_params
from ..hooke.sym import smat_sym_ortho_3d

from .meta import Surface, SurfaceLayer


__all__ = ["Membrane", "MembraneLayer"]


class MembraneLayer(SurfaceLayer):
    """
    Helper object for the stress analysis of a layer of a membrane.
    """

    __loc__ = [-1.0, 1.0]  # locations of approximation points
    __shape__ = (3, 3)  # shape of the ABDS matrix

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # locations of discrete points along the thickness
        self.zi = [self.loc_to_z(loc) for loc in self.__loc__]

    def material_elastic_stiffness_matrix(self) -> ndarray:
        """
        Returns and stores the transformed material stiffness matrix.
        """
        Cm_126 = self.hooke
        T_126 = self.rotation_matrix()
        R_126 = np.diag([1, 1, 2])
        C_126 = T_126 @ Cm_126 @ inv(R_126) @ T_126.T @ R_126
        C_126[np.abs(C_126) < 1e-12] = 0.0
        self.C_126 = C_126
        return C_126

    def rotation_matrix(self) -> ndarray:
        """
        Returns the transformation matrix of the layer.
        """
        angle = self.angle * np.pi / 180
        T_126 = np.zeros([3, 3])
        T_126[0, 0] = np.cos(angle) ** 2
        T_126[0, 1] = np.sin(angle) ** 2
        T_126[0, 2] = -np.sin(2 * angle)
        T_126[1, 0] = T_126[0, 1]
        T_126[1, 1] = T_126[0, 0]
        T_126[1, 2] = -T_126[0, 2]
        T_126[2, 0] = np.cos(angle) * np.sin(angle)
        T_126[2, 1] = -T_126[2, 0]
        T_126[2, 2] = np.cos(angle) ** 2 - np.sin(angle) ** 2
        return T_126

    def elastic_stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness contribution of the layer.
        """
        return self.material_elastic_stiffness_matrix() * (self.tmax - self.tmin)

    def approxfunc(self, data):
        """
        Returns a function that can be used for approximations thorugh
        the thickness.
        """
        z0, z1 = self.zi
        z = np.array([[1, z0], [1, z1]])
        a, b = np.linalg.inv(z) @ np.array(data)
        return lambda z: a + b * z


class Membrane(Surface):
    """
    Helper object for the stress analysis of a membrane.

    Example
    -------
    >>> from sigmaepsilon.model import Membrane as Model
    >>> model = {
    >>>     '0' : {
    >>>         'hooke' : Model.Hooke(E=2100000, nu=0.3),
    >>>         'angle' : 0.,
    >>>         'thickness' : 0.1
    >>>         },
    >>>     }
    >>> C = Model.from_dict(model).stiffness_matrix()
    """

    __layerclass__ = MembraneLayer

    __imap__ = {0: (0, 0), 1: (1, 1), 2: (2, 2), 5: (1, 2), 4: (0, 2), 3: (0, 1)}

    @classmethod
    def Hooke(cls, symbolic: bool = False, **kwargs) -> Union[sy.Matrix, np.ndarray]:
        """
        Returns a Hooke matrix appropriate for membranes.

        Parameters
        ----------
        symbolic : bool, Optional
            If True, a symbolic matrix is returned.

        Returns
        -------
        Union[sy.Matrix, np.ndarray]
            The matrix in symbolic or numerical form.

        Examples
        --------
        >>> from sigmaepsilon.material import Membrane
        >>> Membrane.Hooke(E=2100000, nu=0.3)
        >>> Membrane.Hooke(symbolic=True)
        """
        S = smat_sym_ortho_3d()
        S.row_del(2)
        S.row_del(2)
        S.row_del(2)
        S.col_del(2)
        S.col_del(2)
        S.col_del(2)

        if symbolic:
            return S.inv()

        constants = _get_elastic_params(**kwargs)
        subs = {s: constants[str(s)] for s in S.free_symbols}
        S = S.subs([(sym, val) for sym, val in subs.items()])
        S = np.array(S, dtype=float)
        return np.linalg.inv(S)

    def elastic_stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness matrix of the membrane.

        Returns
        -------
        ndarray
            The ABDS matrix of the membrane.
        """
        self.ABDS = super().elastic_stiffness_matrix()
        self.SDBA = np.linalg.inv(self.ABDS)
        return self.ABDS



import numpy as np
from numba import njit, prange
from collections import Iterable

from neumann import ascont, atleast1d, atleast2d, atleast3d, atleast4d

from .surface import layers_of_points, points_of_layers
from .mindlin import (
    z_to_shear_factors,
    shear_correction_data,
    material_stiffness_matrices,
)

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D(
    ABDS: np.ndarray,
    sfx: np.ndarray,
    sfy: np.ndarray,
    C_126: np.ndarray,
    C_45: np.ndarray,
    bounds: np.ndarray,
    points: np.ndarray,
    e_126_n: np.ndarray,
    c_126: np.ndarray,
    e_45: np.ndarray,
):
    nRHS, nP, _ = e_126_n.shape
    layerinds = layers_of_points(points, bounds)

    # results
    s_126_n = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nRHS, nP, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    e_45_new = np.zeros((nRHS, nP, 2), dtype=ABDS.dtype)

    for iP in prange(nP):
        zP = points[iP, 2]
        lP = layerinds[iP]
        sfxz, sfyz = z_to_shear_factors(zP, sfx[lP, :], sfy[lP, :])
        for iRHS in prange(nRHS):
            e_126_m[iRHS, iP] = zP * c_126[iRHS, iP]
            s_126_n[iRHS, iP] = C_126[lP] @ e_126_n[iRHS, iP]
            s_126_m[iRHS, iP] = C_126[lP] @ e_126_m[iRHS, iP]
            s_45[iRHS, iP, 0] = sfxz * ABDS[6, 6] * e_45[iRHS, iP, 0]
            s_45[iRHS, iP, 1] = sfyz * ABDS[7, 7] * e_45[iRHS, iP, 1]
            e_45_new[iRHS, iP, 0] = s_45[iRHS, iP, 0] / C_45[lP, 0, 0]
            e_45_new[iRHS, iP, 1] = s_45[iRHS, iP, 1] / C_45[lP, 1, 1]

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D_s(
    ABDS: np.ndarray,
    sfx: np.ndarray,
    sfy: np.ndarray,
    C_126: np.ndarray,
    C_45: np.ndarray,
    bounds: np.ndarray,
    ppl: int,
    e_126_n: np.ndarray,
    c_126: np.ndarray,
    e_45: np.ndarray,
):
    nRHS, nP, _ = e_126_n.shape
    z = points_of_layers(bounds, ppl)
    nL = z.shape[0]

    # results
    s_126_n = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nRHS, nP, nL, ppl, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    e_45_new = np.zeros((nRHS, nP, nL, ppl, 2), dtype=ABDS.dtype)

    for iL in prange(nL):
        for ippl in prange(ppl):
            zP = z[iL, ippl]
            sfxz, sfyz = z_to_shear_factors(zP, sfx[iL, :], sfy[iL, :])
            for iP in prange(nP):
                for iRHS in prange(nRHS):
                    e_126_m[iRHS, iP, iL, ippl] = zP * c_126[iRHS, iP]
                    s_126_n[iRHS, iP, iL, ippl] = C_126[iL] @ e_126_n[iRHS, iP]
                    s_126_m[iRHS, iP, iL, ippl] = C_126[iL] @ e_126_m[iRHS, iP]
                    s_45[iRHS, iP, iL, ippl, 0] = sfxz * ABDS[6, 6] * e_45[iRHS, iP, 0]
                    s_45[iRHS, iP, iL, ippl, 1] = sfyz * ABDS[7, 7] * e_45[iRHS, iP, 1]
                    e_45_new[iRHS, iP, iL, ippl, 0] = (
                        s_45[iRHS, iP, iL, ippl, 0] / C_45[iL, 0, 0]
                    )
                    e_45_new[iRHS, iP, iL, ippl, 1] = (
                        s_45[iRHS, iP, iL, ippl, 1] / C_45[iL, 1, 1]
                    )

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D_rgrid(
    ABDS: np.ndarray,
    coords2d: np.ndarray,
    topo2d: np.ndarray,
    bounds: np.ndarray,
    cell_per_layer: int,
    eshape: tuple,
    sfx: np.ndarray,
    sfy: np.ndarray,
    C_126: np.ndarray,
    C_45: np.ndarray,
    e_126_n: np.ndarray,
    c_126: np.ndarray,
    e_45: np.ndarray,
):
    nNEx, nNEy, nNEz = eshape
    nNE3d = nNEx * nNEy * nNEz
    nRHS = e_126_n.shape[0]
    nL = bounds.shape[0]
    nE2d, nNE2d = topo2d.shape
    nE3d = nE2d * nL * cell_per_layer
    nN3d = nE3d * nNE3d

    ftype = coords2d.dtype
    itype = topo2d.dtype
    s_126_n = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    s_126_m = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    s_45 = np.zeros((nRHS, nN3d, 2), dtype=ftype)
    e_126_m = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    e_45_new = np.zeros((nRHS, nN3d, 2), dtype=ftype)
    coords3d = np.zeros((nN3d, 3), dtype=ftype)
    topo3d = np.zeros((nE3d, nNE3d), dtype=itype)

    # precalculate shear factors
    sfxz = np.zeros((nL, cell_per_layer, nNEz), dtype=ftype)
    sfyz = np.zeros((nL, cell_per_layer, nNEz), dtype=ftype)
    for iL in prange(nL):
        dZ = (bounds[iL, 1] - bounds[iL, 0]) / cell_per_layer
        ddZ = dZ / (nNEz - 1)
        for iCL in prange(cell_per_layer):
            for iNEz in prange(nNEz):
                zP = dZ * iCL + ddZ * iNEz
                sfxz[iL, iCL, iNEz], sfyz[iL, iCL, iNEz] = z_to_shear_factors(
                    zP, sfx[iL, :], sfy[iL, :]
                )

    for iL in prange(nL):
        dZ = (bounds[iL, 1] - bounds[iL, 0]) / cell_per_layer
        ddZ = dZ / (nNEz - 1)
        for iCL in prange(cell_per_layer):
            for iE2d in prange(nE2d):
                iE3d = nE2d * (cell_per_layer * iL + iCL) + iE2d
                for iNEx in prange(nNEx):
                    for iNEy in prange(nNEy):
                        iNE2d = iNEx * nNEy + iNEy
                        iN2d = topo2d[iE2d, iNE2d]
                        for iNEz in prange(nNEz):
                            iNE3d = iNEx * iNEy * nNEz + iNEz
                            iN3d = iE3d * nNE3d + iNE3d
                            topo3d[iE3d, iNE3d] = iN3d
                            coords3d[iN3d, 0] = coords2d[iN2d, 0]
                            coords3d[iN3d, 1] = coords2d[iN2d, 1]
                            coords3d[iN3d, 2] = dZ * iCL + ddZ * iNEz
                            for iRHS in prange(nRHS):
                                e_126_m[iRHS, iN3d] = (
                                    coords3d[iN3d, 2] * c_126[iRHS, iN2d]
                                )
                                s_126_n[iRHS, iN3d] = C_126[iL] @ e_126_n[iRHS, iN2d]
                                s_126_m[iRHS, iN3d] = C_126[iL] @ e_126_m[iRHS, iN3d]
                                s_45[iRHS, iN3d, 0] = (
                                    sfxz[iL, iCL, iNEz]
                                    * ABDS[6, 6]
                                    * e_45[iRHS, iN2d, 0]
                                )
                                s_45[iRHS, iN3d, 1] = (
                                    sfyz[iL, iCL, iNEz]
                                    * ABDS[7, 7]
                                    * e_45[iRHS, iN2d, 1]
                                )
                                e_45_new[iRHS, iN3d, 0] = (
                                    s_45[iRHS, iN3d, 0] / C_45[iL, 0, 0]
                                )
                                e_45_new[iRHS, iN3d, 1] = (
                                    s_45[iRHS, iN3d, 1] / C_45[iL, 1, 1]
                                )

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


def pproc_Mindlin_3D(
    ABDS: np.ndarray,
    C_126: Iterable,
    C_45: Iterable,
    bounds: Iterable,
    points: np.ndarray,
    strains2d: np.ndarray,
    *args,
    angles: Iterable = None,
    separate=True,
    shear_factors: Iterable = None,
    **kwargs
):
    # formatting
    ABDS = atleast3d(ABDS)
    strains2d = atleast4d(strains2d)
    if isinstance(C_126, np.ndarray):
        C_126 = atleast4d(C_126)
    else:
        C_126 = [atleast3d(C_126[i]) for i in range(len(C_126))]
    if isinstance(C_45, np.ndarray):
        C_45 = atleast4d(C_45)
    else:
        C_45 = [atleast3d(C_45[i]) for i in range(len(C_45))]
    if isinstance(bounds, np.ndarray):
        bounds = atleast3d(bounds)
    else:
        bounds = [atleast2d(bounds[i]) for i in range(len(bounds))]

    # transform material stiffness matrices to global
    if angles is not None:
        C_126_g = np.zeros_like(C_126)
        C_45_g = np.zeros_like(C_45)
        if isinstance(angles, np.ndarray):
            angles = atleast2d(angles)
        else:
            angles = [atleast1d(angles[i]) for i in range(len(angles))]
        for iLHS in range(len(angles)):
            C_126_g[iLHS], C_45_g[iLHS] = material_stiffness_matrices(
                C_126[iLHS], C_45[iLHS], angles[iLHS]
            )
        C_126 = C_126_g
        C_45 = C_45_g
        del C_126_g
        del C_45_g

    # calculate shear factors
    if shear_factors is None:
        shear_factors = []
        for i in range(len(C_126)):
            _, sf = shear_correction_data(ABDS[i], C_126[i], C_45[i], bounds[i])
            shear_factors.append(sf)
    else:
        if isinstance(shear_factors, np.ndarray):
            shear_factors = atleast4d(shear_factors)
        else:
            shear_factors = [
                atleast3d(shear_factors[i]) for i in range(len(shear_factors))
            ]

    nLHS, nRHS, nP, _ = strains2d.shape
    e_126_n = ascont(strains2d[:, :, :, :3])
    s_126_n = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nLHS, nRHS, nP, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    e_45 = np.zeros((nLHS, nRHS, nP, 2), dtype=ABDS.dtype)

    for i in range(nLHS):
        sfx_i = ascont(shear_factors[i][:, 0, :])
        sfy_i = ascont(shear_factors[i][:, 1, :])
        e_126_n_i = ascont(strains2d[i, :, :, :3])
        c_126_i = ascont(strains2d[i, :, :, 3:6])
        e_45_i = ascont(strains2d[i, :, :, 6:8])
        (
            s_126_n[i],
            s_126_m[i],
            s_45[i],
            e_126_n[i],
            e_126_m[i],
            e_45[i],
        ) = _pproc_Mindlin_3D(
            ABDS[i],
            sfx_i,
            sfy_i,
            C_126[i],
            C_45[i],
            bounds[i],
            points,
            e_126_n_i,
            c_126_i,
            e_45_i,
        )

    if separate:
        res3d = s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45
    else:
        res3d = s_126_n + s_126_m, s_45, e_126_n + e_126_m, e_45

    return res3d



from polymesh.cells import TET4 as Tetra

from ..material.solid3d import Solid3d
from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class TET4(ABC, Solid3d, Tetra, FiniteElement):
    qrule = "full"



from .fem.mesh import FemMesh, SolidMesh
from .fem.pointdata import PointData
from .fem.structure import Structure
from .fem.linemesh import LineMesh
from .fem.surfacemesh import SurfaceMesh
from .material.beam.bernoulli.section import BeamSection

__version__ = "0.0.32"

__description__ = "High-Performance Computational Mechanics in Python."



from typing import Tuple

import numpy as np
from numpy import ndarray
from scipy.optimize import minimize

from dewloosh.core.tools.kwargtools import getasany
from neumann.function import Function
from neumann.linalg import is_pos_def

from .sym import smat_sym_ortho_3d

__all__ = ["elastic_compliance_matrix", "elastic_stiffness_matrix"]


poisson_ratios = ["NU12", "NU13", "NU23", "NU21", "NU32", "NU31"]
elastic_moduli = ["E1", "E2", "E3"]
shear_moduli = ["G12", "G13", "G23"]
standard_keys_ortho = elastic_moduli + poisson_ratios[:3] + shear_moduli
keys_ortho_all = standard_keys_ortho + poisson_ratios[3:]


class HookeError(Exception):
    ...


def _find_all_ortho_params(**params):

    # select 9 parameters (3 shear moduli and 6 from the others)
    _params = dict()
    Young, Poisson, Shear = _group_params(**params)
    _params.update(Shear)
    Young.update(Poisson)
    for i, (k, v) in enumerate(Young.items()):
        if i < 6:
            _params[k] = v

    S = smat_sym_ortho_3d()
    S_skew = S - S.T
    subs = [(sym, val) for sym, val in _params.items() if val is not None]
    residual = S_skew.subs(subs)
    error = residual.norm()
    f = Function(error)
    v = f.variables
    x0 = np.zeros((len(v),), dtype=float)
    res = minimize(f, x0, method="Nelder-Mead", tol=1e-6)
    _params.update({str(sym): val for sym, val in zip(v, res.x)})
    return _params


def _group_params(**params) -> tuple:
    """
    Groups and returns all input material parameters as a 3-tuple
    of dictionaries.

    Returns
    -------
    tuple
        3 dictionaries for Young's moduli, Poisson's ratios
        and shear moduli.
    """
    PARAMS = {key.upper(): value for key, value in params.items()}
    keys = list(PARAMS.keys())
    E = {key: PARAMS[key] for key in keys if key[0] == "E"}
    NU = {key: PARAMS[key] for key in keys if key[:2] == "NU"}
    G = {key: PARAMS[key] for key in keys if key[0] == "G"}
    return E, NU, G


def _missing_std_params_ortho(**params):
    """
    Returns the ones missing from the 9 standard orthotropic parameters
    E1, E2, E3, NU12, NU13, NU23, G12, G13, and G23.
    """
    _params = [k for k, v in params.items() if v is not None]
    return [k for k in standard_keys_ortho if k not in _params]


def _missing_params_ortho(**params):
    """
    Returns the ones missing from the 12 orthotropic parameters
    E1, E2, E3, NU12, NU13, NU23, G12, G13, and G23.
    """
    _params = [k for k, v in params.items() if v is not None]
    return [k for k in keys_ortho_all if k not in _params]


def _has_std_params_ortho(**params) -> bool:
    """
    Returns True, if all 9 standard parameters E1, E2, E3, NU12, NU13, NU23,
    G12, G13, and G23 are provided.
    """
    return len(_missing_std_params_ortho(**params)) == 0


def _has_all_params_ortho(**params) -> bool:
    """
    Returns True, if all 12 keys of an orthotropic material are provided.
    """
    return len(_missing_params_ortho(**params)) == 0


def _finalize_iso(E: float = None, NU: float = None, G: float = None) -> Tuple[float]:
    try:
        if E is None:
            E = 2 * G * (1 + NU)
        elif G is None:
            G = E / 2 / (1 + NU)
        elif NU is None:
            NU = E / 2 / G - 1
        return E, NU, G
    except TypeError:
        msg = (
            "At least 2 independent constants",
            " must be defined for an isotropic material",
        )
        raise HookeError(msg)


def _finalize_ortho(**params):
    if _has_std_params_ortho(**params):
        params["NU21"] = params["NU12"] * params["E2"] / params["E1"]
        params["NU31"] = params["NU13"] * params["E3"] / params["E1"]
        params["NU32"] = params["NU23"] * params["E3"] / params["E2"]
        return params
    else:
        E, NU, G = _group_params(**params)
        nE, nNU, nG = len(E), len(NU), len(G)
        
        if not nG == 3:
            msg = "The three shear moduli must be provided!"
            raise HookeError(msg)

        if not (nE + nNU) >= 6:
            msg = (
                "6 from the 3 Young's moduli and the 6 Poisson's"
                " ratios must be provided!"
            )
            raise HookeError(msg)

        params = _find_all_ortho_params(**params)
        assert _has_all_params_ortho(**params), "Something went wrong!"


def _get_iso_params(**kwargs):
    """
    Returns all 12 orthotropic engineering constants for an isotropic material.

    Requires 2 independent constants from E, NU and G to be provided.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}
    E = KWARGS.get("E", None)
    G = KWARGS.get("G", None)
    NU = KWARGS.get("NU", None)
    E, NU, G = _finalize_iso(E, NU, G)
    params = {
        "E1": E,
        "E2": E,
        "E3": E,
        "NU12": NU,
        "NU13": NU,
        "NU23": NU,
        "NU21": NU,
        "NU31": NU,
        "NU32": NU,
        "G12": G,
        "G13": G,
        "G23": G,
    }
    return params


def _get_triso_params_23(**kwargs) -> dict:
    """
    Returns all 12 orthotropic engineering constants for a
    transversely isotropic material.

    From the total of 12 constants, exactly 5 must be provided,
    from which 1 must be the out-of-plane shear modulus. The remaining 4
    necessary constants can be provided in any possible combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}

    G13 = getasany(["G13", "G31", "G12", "G21"], None, **KWARGS)
    assert G13 is not None, "The out-of-plane shear modulus must be provided!"

    NU12 = getasany(["NU12", "NU13"], None, **KWARGS)
    NU21 = getasany(["NU21", "NU31"], None, **KWARGS)
    NU23 = getasany(["NU23", "NU32"], None, **KWARGS)
    E2 = getasany(["E2", "E3"], None, **KWARGS)
    E1 = KWARGS.get("E1", None)
    G23 = getasany(["G23", "G32"], None, **KWARGS)

    E2, NU23, G23 = _finalize_iso(E2, NU23, G23)

    # from material symmetry
    NU32 = NU23
    NU13 = NU12
    NU31 = NU21
    E3 = E2
    G12 = G13

    params = {
        "E1": E1,
        "E2": E2,
        "E3": E3,
        "NU12": NU12,
        "NU13": NU13,
        "NU23": NU23,
        "NU21": NU21,
        "NU31": NU31,
        "NU32": NU32,
        "G12": G12,
        "G13": G13,
        "G23": G23,
    }
    return _finalize_ortho(**params)


def _get_triso_params_13(**kwargs) -> dict:
    """
    Returns all 12 orthotropic engineering constants for a
    transversely isotropic material.

    From the total of 12 constants, exactly 5 must be provided,
    from which 1 must be the out-of-plane shear modulus. The remaining 4
    necessary constants can be provided in any possible combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}

    G12 = getasany(["G12", "G21", "G23", "G32"], None, **KWARGS)
    assert G12 is not None, "The out-of-plane shear modulus must be provided!"

    NU12 = getasany(["NU12", "NU32"], None, **KWARGS)
    NU21 = getasany(["NU21", "NU23"], None, **KWARGS)
    NU13 = getasany(["NU13", "NU31"], None, **KWARGS)
    E1 = getasany(["E1", "E3"], None, **KWARGS)
    E2 = KWARGS.get("E2", None)
    G13 = getasany(["G13", "G31"], None, **KWARGS)

    E1, NU13, G13 = _finalize_iso(E1, NU13, G13)

    # from material symmetry
    NU31 = NU13
    NU32 = NU12
    NU23 = NU21
    E3 = E1
    G23 = G12

    params = {
        "E1": E1,
        "E2": E2,
        "E3": E3,
        "NU12": NU12,
        "NU13": NU13,
        "NU23": NU23,
        "NU21": NU21,
        "NU31": NU31,
        "NU32": NU32,
        "G12": G12,
        "G13": G13,
        "G23": G23,
    }
    return _finalize_ortho(**params)


def _get_triso_params_12(**kwargs) -> dict:
    """
    Returns all 12 orthotropic engineering constants for a
    transversely isotropic material.

    From the total of 12 constants, exactly 5 must be provided,
    from which 1 must be the out-of-plane shear modulus. The remaining 4
    necessary constants can be provided in any possible combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}

    G13 = getasany(["G13", "G31", "G23", "G32"], None, **KWARGS)
    assert G13 is not None, "The out-of-plane shear modulus must be provided!"

    NU13 = getasany(["NU13", "NU23"], None, **KWARGS)
    NU31 = getasany(["NU31", "NU32"], None, **KWARGS)
    NU12 = getasany(["NU12", "NU21"], None, **KWARGS)
    E1 = getasany(["E1", "E2"], None, **KWARGS)
    E3 = KWARGS.get("E3", None)
    G12 = getasany(["G12", "G21"], None, **KWARGS)

    E1, NU12, G12 = _finalize_iso(E1, NU12, G12)

    # from material symmetry
    NU21 = NU12
    NU23 = NU13
    NU32 = NU31
    E2 = E1
    G23 = G13

    params = {
        "E1": E1,
        "E2": E2,
        "E3": E3,
        "NU12": NU12,
        "NU13": NU13,
        "NU23": NU23,
        "NU21": NU21,
        "NU31": NU31,
        "NU32": NU32,
        "G12": G12,
        "G13": G13,
        "G23": G23,
    }
    return _finalize_ortho(**params)


def _get_ortho_params(**kwargs) -> dict:
    """
    Returns all 12 engineering constants for an orthotropic material.

    From the total of 12 constants, 9 must be provided, from which
    3 must be shear moduli. The remaining 6 necessary constants
    can be provided in any combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}
    E, NU, G = _group_params(**kwargs)
    nE, nNU, nG = len(E), len(NU), len(G)

    if not nG == 3:
        msg = "The three shear moduli must be provided!"
        raise HookeError(msg)

    if not (nE + nNU) >= 6:
        msg = (
            "6 from the 3 Young's moduli and the 6 Poisson's"
            " ratios must be provided!"
        )
        raise HookeError(msg)

    G12 = getasany(["G12", "G21"], None, **KWARGS)
    if G12 is None:
        msg = "Shear modulus G12 must be provided as either G12 or G21!"
        raise HookeError(msg)

    G23 = getasany(["G23", "G32"], None, **KWARGS)
    if G23 is None:
        msg = "Shear modulus G23 must be provided as either G23 or G32!"
        raise HookeError(msg)

    G13 = getasany(["G13", "G31"], None, **KWARGS)
    if G13 is None:
        msg = "Shear modulus G13 must be provided as either G13 or G31!"
        raise HookeError(msg)

    E1 = KWARGS.get("E1", None)
    E2 = KWARGS.get("E2", None)
    E3 = KWARGS.get("E3", None)
    NU12 = KWARGS.get("NU12", None)
    NU21 = KWARGS.get("NU21", None)
    NU23 = KWARGS.get("NU23", None)
    NU32 = KWARGS.get("NU32", None)
    NU13 = KWARGS.get("NU13", None)
    NU31 = KWARGS.get("NU31", None)

    params = {
        "E1": E1,
        "E2": E2,
        "E3": E3,
        "NU12": NU12,
        "NU13": NU13,
        "NU23": NU23,
        "NU21": NU21,
        "NU31": NU31,
        "NU32": NU32,
        "G12": G12,
        "G13": G13,
        "G23": G23,
    }
    return _finalize_ortho(**params)


def _get_elastic_params(isoplane: str = None, **params):
    E, NU, G = _group_params(**params)
    nE, nNU, nG = len(E), len(NU), len(G)
    nParams = sum([nE, nNU, nG])

    if nParams == 2:
        params = _get_iso_params(**E, **NU, **G)
    elif nParams == 5:
        if not isinstance(isoplane, str):
            raise ValueError(
                "The plane of isotropy must be specified. See the docs!")
        if isoplane in ["23", "32"]:
            params = _get_triso_params_23(**E, **NU, **G)
        elif isoplane in ["12", "21"]:
            params = _get_triso_params_12(**E, **NU, **G)
        elif isoplane in ["13", "31"]:
            params = _get_triso_params_13(**E, **NU, **G)
    elif nParams == 9:
        params = _get_ortho_params(**E, **NU, **G)
    else:
        raise HookeError("Invalid material definition!")

    return params


def elastic_compliance_matrix(
    imap: dict = None, verify: bool = False, **params
) -> ndarray:
    """
    Returns the elastic compliance matrix for a Hooke-type material model from
    a set of engineering constants.

    Parameters
    ----------
    imap: dict, Optional
        Index mapping. Default is the Voigt-map.
    verify: bool, Optional
        If True, the resulting matrix is verified before getting returned. Default is False.
    isoplane: str, Optional
        If the number of specified engineering constants is 5, it suggests a transversely
        orthotropic material. In this case, the parameter 'isoplane' defines the plane of
        isotropy. Valid choices are '12', '23' and '13'. Default is None.
        
    Raises
    ------
    ~sigmaepsilon.material.hooke.utils.HookeError

    Returns
    -------
    numpy.ndarray
        A 6x6 NumPy float array.
    """
    params = _get_elastic_params(**params)
    sympy_matrix = smat_sym_ortho_3d(imap=imap, **params)
    S = np.array(sympy_matrix.tolist()).astype(float)
    if verify:
        if not is_pos_def(S):
            raise HookeError(
                "The parameters does not result in a positive definite material!")
    return S


def elastic_stiffness_matrix(*args, verify: bool = False, **kwargs) -> ndarray:
    """
    Returns the elastic stiffness matrix for a Hooke-type material model from
    a set of engineering constants.

    Parameters
    ----------
    imap: dict, Optional
        Index mapping. Default is the Voigt-map.
    verify: bool, Optional
        If True, the resulting matrix is verified before getting returned. Default is False.
    isoplane: str, Optional
        If the number of specified engineering constants is 5, it suggests a transversely
        orthotropic material. In this case, the parameter 'isoplane' defines the plane of
        isotropy. Valid choices are '12', '23' and '13'. Default is None.
        
    Raises
    ------
    ~sigmaepsilon.material.hooke.utils.HookeError

    Returns
    -------
    numpy.ndarray
        A 6x6 NumPy float array.
    """
    C = np.linalg.inv(elastic_compliance_matrix(*args, **kwargs))
    if verify:
        if not is_pos_def(C):
            raise HookeError(
                "The parameters does not result in a positive definite material!")
    return C



from numba import njit, prange
import numpy as np
from numpy import ndarray

from .solid import Solid
from ...utils.material.hmh import HMH_3d

__cache = True


__all__ = ["Solid3d"]


_NSTRE_ = 6
_NDOFN_ = 3


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(dshp: ndarray, jac: ndarray) -> ndarray:
    # sxx, syy, szz, syz, sxz, sxy
    nE = jac.shape[0]
    nP, nN = dshp.shape[:2]
    nTOTV = nN * _NDOFN_
    B = np.zeros((nE, nP, _NSTRE_, nTOTV), dtype=dshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            gdshp = dshp[iP] @ np.linalg.inv(jac[iE, iP]).T
            for i in prange(nN):
                B[iE, iP, 0, 0 + i * _NDOFN_] = gdshp[i, 0]
                B[iE, iP, 1, 1 + i * _NDOFN_] = gdshp[i, 1]
                B[iE, iP, 2, 2 + i * _NDOFN_] = gdshp[i, 2]
                B[iE, iP, 3, 1 + i * _NDOFN_] = gdshp[i, 2]
                B[iE, iP, 3, 2 + i * _NDOFN_] = gdshp[i, 1]
                B[iE, iP, 4, 0 + i * _NDOFN_] = gdshp[i, 2]
                B[iE, iP, 4, 2 + i * _NDOFN_] = gdshp[i, 0]
                B[iE, iP, 5, 0 + i * _NDOFN_] = gdshp[i, 1]
                B[iE, iP, 5, 1 + i * _NDOFN_] = gdshp[i, 0]
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def HMH_3d_bulk_multi(estrs: np.ndarray) -> ndarray:
    nE, nP = estrs.shape[:2]
    res = np.zeros((nE, nP), dtype=estrs.dtype)
    for iE in prange(nE):
        for jNE in prange(nP):
            res[iE, jNE] = HMH_3d(estrs[iE, jNE])
    return res


class Solid3d(Solid):
    """
    A model for 3d solids.

    See Also
    --------
    :class:`sigmaepsilon.fem.model.solid.Solid`
    """

    dofs = ("UX", "UY", "UZ")

    NDOFN = 3
    NSTRE = 6

    def strain_displacement_matrix(
        self, pcoords: ndarray = None, *, dshp: ndarray = None, jac: ndarray = None, **_
    ) -> ndarray:
        """
        Calculates the strain displacement matrix for solids.

        Parameters
        ----------
        pcoords : numpy.ndarray, Optional
            Locations of evaluation points. Either 'pcoords' or
            both 'dshp' and 'jac' must be provided.
        dshp : numpy.ndarray, Optional
            Shape function derivatives evaluated at some points.
            Only required if 'pcoords' is not provided.
        jac : numpy.ndarray
            Jacobian matrices evaluated at some points.
            Only required if 'pcoords' is not provided.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nP, nSTRE, nTOTV), where nE, nP
            nSTRE and nTOTV are the number of elements, evaulation
            points, stress components and total DOFs per element.
        """
        if dshp is None and pcoords is not None:
            dshp = self.shape_function_derivatives(pcoords)
        if jac is None:
            jac = self.jacobian_matrix(dshp=dshp)
        return strain_displacement_matrix(dshp, jac)

    @classmethod
    def HMH(cls, stresses: ndarray) -> ndarray:
        """
        Evaluates the Huber-Mises-Hencky stress at multiple points
        of multiple cells.

        s11, s22, s33, s12, s13, s23

        Parameters
        ----------
        stresses : numpy.ndarray, Optional
            Array of shape (nE, nP, nSTRE), where nE, nP and nSTRE
            are the number of elements, evaulation points and stress
            components. The stresses are expected in the order
            s11, s22, s33, s12, s13, s23.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nP), where nE and nP are the
            number of elements and evaulation points.
        """
        return HMH_3d_bulk_multi(stresses)



from polymesh.cells import W6 as Wedge6, W18 as Wedge18

from ..material.solid3d import Solid3d
from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class W6(ABC, Solid3d, Wedge6, FiniteElement):
    qrule = "full"


class W18(ABC, Solid3d, Wedge18, FiniteElement):
    qrule = "full"



from linkeddeepdict import LinkedDeepDict

from .elem import FiniteElement
from .meta import ABCFiniteElement
from .celldata import CellData

from .bernoulli2 import Bernoulli2 as B2
from .bernoulli3 import Bernoulli3 as B3

from .cst import CST_M, CST_P_MR, CST_S_MR
from .lst import LST_M, LST_P_MR, LST_S_MR

from .q4 import Q4_M, Q4_P_MR, Q4_S_MR
from .q9 import Q9_M, Q9_P_MR, Q9_S_MR

from .tet4 import TET4
from .tet10 import TET10

from .h8 import H8
from .h27 import H27

from .wedge import W6, W18

cells_dict = LinkedDeepDict(
    {
        "bernoulli": {"B2": B2, "B3": B3},
        "membrane": {
            "T3": CST_M,
            "T6": LST_M,
            "Q4": Q4_M,
            "Q9": Q9_M,
        },
        "mindlin": {"T3": CST_P_MR, "T6": LST_P_MR, "Q4": Q4_P_MR, "Q9": Q9_P_MR},
        "3d": {"H8": H8, "H27": H27, "TET4": TET4, "TET10": TET10},
    }
)



import numpy as np
from numpy import ndarray
from numba import njit, vectorize, prange

__cache = True


@njit(nogil=True, cache=__cache)
def HMH_M(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for membranes.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s12.
    """
    s11, s22, s12 = strs
    return np.sqrt(s11**2 - s11 * s22 + s22**2 + 3 * s12**2)


@njit(nogil=True, cache=__cache)
def HMH_S(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for shells.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s12, s13, s23.
    """
    s11, s22, s12, s13, s23 = strs
    return np.sqrt(
        s11**2 - s11 * s22 + s22**2 + 3 * s12**2 + 3 * s13**2 + 3 * s23**2
    )


@njit(nogil=True, cache=__cache)
def HMH_3d(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for 3d solids.

    Parameters
    ----------
    strs : numpy.ndarray
        The stresses s11, s22, s33, s23, s13, s12.
    """
    s11, s22, s33, s23, s13, s12 = strs
    return np.sqrt(
        0.5 * ((s11 - s22) ** 2 + (s22 - s33) ** 2 + (s33 - s11) ** 2)
        + 3 * (s12**2 + s13**2 + s23**2)
    )


@njit(nogil=True, cache=__cache)
def HMH_3d_multi(strs: ndarray):
    """
    Evaluates the Huber-Mises-Hencky formula for 3d solids for
    multiple points.

    Parameters
    ----------
    strs : numpy.ndarray
        2d array of stresses for several points. The stresses are
        expected in the order s11, s22, s33, s23, s13, s12.
    """
    nP = strs.shape[0]
    res = np.zeros(nP, dtype=strs.dtype)
    for i in prange(nP):
        res[i] = HMH_3d(strs[i])
    return res


@vectorize("f8(f8, f8, f8, f8, f8, f8)", target="parallel", cache=__cache)
def HMH_3d_v(s11, s22, s33, s23, s13, s12):
    """
    Vectorized evaluation of the HMH failure criterion.

    The input values s11, s22, s33, s23, s13, s12 can be
    arbitrary dimensional arrays.
    """
    return np.sqrt(
        0.5 * ((s11 - s22) ** 2 + (s22 - s33) ** 2 + (s33 - s11) ** 2)
        + 3 * (s12**2 + s13**2 + s23**2)
    )



from typing import Tuple

import numpy as np
from numpy import ndarray
from numba import njit, prange

from polymesh.utils.space import index_of_closest_point
from polymesh.space import PointCloud
from polymesh.cell import PolyCell3d

__cache = True


def link_points_to_points(
    source: PointCloud,
    target: PointCloud,
) -> ndarray:
    i = index_of_closest_point(source.show(), target.show())
    link = np.stack([source.id[i], target.id], axis=1)
    return link


def link_points_to_body(
    points: PointCloud, 
    body: PolyCell3d,
    lazy: bool = True,
    tol: float = 1e-12,
    k: int = 4,
) -> Tuple[ndarray]:
    source_coords = points.show()
    source_indices = points.id
    topo_target = body.topology()
        
    i_source, i_target, locations_target = \
        body.locate(source_coords, lazy, tol, k)
    source_indices = source_indices[i_source]
    topo_target = topo_target[i_target]
    
    shp_target = body.shape_function_values(locations_target)
            
    nE, nNE = topo_target.shape
    factors = np.zeros((nE, nNE + 1), dtype=float)
    indices = np.zeros((nE, nNE + 1), dtype=int)
    factors[:, -1] = -1
    factors[:, :nNE] = shp_target
    indices[:, -1] = source_indices
    indices[:, :nNE] = topo_target
    
    return factors, indices


@njit(nogil=True, parallel=True, cache=__cache)
def _body_to_body_stiffness_data_(
    nodal_factors:ndarray, nodal_indices:ndarray, dofmap:ndarray, ndof:int
) -> Tuple[ndarray]:
    nN, nV = nodal_factors.shape
    nDOF = dofmap.shape[0]
    nR = nN * nDOF
    factors = np.zeros((nR, nV), dtype=nodal_factors.dtype)
    indices = np.zeros((nR, nV), dtype=nodal_indices.dtype)
    for i in prange(nN):
        for j in prange(nDOF):
            ii = i*nDOF + j
            for k in prange(nV):
                ik = nodal_indices[i, k] * ndof + dofmap[j]
                factors[ii, k] = nodal_factors[i, k]
                indices[ii, k] = ik
    return factors, indices


import numpy as np
from typing import Union, Iterable

from linkeddeepdict import LinkedDeepDict
from neumann import atleast1d

from .problem import NavierProblem
from .loads import LoadGroup, NavierLoadError
from .preproc import lhs_Navier, rhs_Bernoulli
from .postproc import postproc
from .proc import linsolve_Bernoulli, linsolve_Timoshenko


class NavierBeam(NavierProblem):
    """
    A class to handle simply-supported plates and their solution
    using Navier's approach.

    Parameters
    ----------
    length : float
        The length of the beam.
    N : int, Optional
        The number of harmonic terms involved in the approximation.
        Default is 100.
    EI : float
        Bending stiffness. Default is None.
    GA : float, Optional
        Shear stiffness. Only for Timoshenko beams. Default is None.

    Examples
    --------
    To define an Euler-Bernoulli beam of length 10.0 and
    bending stiffness 2000.0 with 100 harmonic terms involved:

    >>> from sigmaepsilon.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0)

    To define a Timoshenko beam of length 10.0, bending stiffness
    2000.0 and shear stiffness 1500.0 with 100 harmonic terms involved:

    >>> from sigmaepsilon.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0, GA=1500.0)

    """

    def __init__(
        self, length: float, N: int = 100, *, EI: float = None, GA: float = None
    ):
        super().__init__()
        self.length = length
        self.EI = EI
        self.GA = GA
        self.N = N

    def solve(
        self, loads: Union[dict, LoadGroup], points: Union[float, Iterable]
    ) -> LinkedDeepDict:
        """
        Solves the problem and calculates all postprocessing quantities at
        one ore more points.

        Parameters
        ----------
        loads : Union[dict, LoadGroup]
            The loads. If it is an array, it should be a 2d float array.
        points : float or Iterable
            A float or an 1d iterable of coordinates, where the results are
            to be evaluated. If it is a scalar, the resulting dictionary
            contains 1d arrays for every quantity, for every load case. If
            there are multiple points, the result attached to a load case is
            a 2d array, where the first axis goes along the points.

        Returns
        -------
        LinkedDeepDict
            A nested dictionary with a same layout as the input.
        """
        # STIFFNESS
        LHS = lhs_Navier(self.length, self.N, D=self.EI, S=self.GA)

        # LOADS
        if isinstance(loads, LoadGroup):
            _loads = loads
        elif isinstance(loads, dict):
            _loads = LoadGroup.from_dict(loads)
        else:
            raise NavierLoadError()
        _loads.problem = self
        LC = list(_loads.cases())
        RHS = np.vstack(list(lc.rhs() for lc in LC))

        # SOLUTION
        if self.GA is None:
            _RHS = rhs_Bernoulli(RHS, self.length)
            coeffs = linsolve_Bernoulli(LHS, _RHS)
            del _RHS
            # (nLHS, nRHS, nMN)
        else:
            coeffs = linsolve_Timoshenko(LHS, RHS)
            # (nLHS, nRHS, nMN, 2)

        # POSTPROCESSING
        points = atleast1d(points)
        res = postproc(self.length, self.N, points, coeffs, RHS, self.EI, self.GA)
        # (nLHS, nRHS, nP, nX)
        result = LinkedDeepDict()
        for i, lc in enumerate(LC):
            result[lc.address] = res[0, i, :, :]
        result.lock()
        return result



from numpy import ndarray

from neumann import atleast2d, atleast3d

from ...utils.material import model_strains, model_strains_multi, stresses_from_strains
from ...utils.fem.fem import topo_to_gnum
from ..cells.meta import FemMaterial


__all__ = ["Solid"]


class Solid(FemMaterial):
    dofs = ("UX", "UY", "UZ")
    strn = ("exx", "eyy", "ezz", "eyz", "exz", "exy")

    NDOFN = 3
    NSTRE = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def elastic_material_stiffness_matrix(self) -> ndarray:
        return self.material_stiffness

    @classmethod
    def HMH(cls, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def model_strains(cls, dofsol1d, gnum, B, *args, **kwargs):
        if len(dofsol1d.shape) == 2:
            return model_strains_multi(dofsol1d, gnum, B)
        return model_strains(dofsol1d, gnum, B)

    def strains_at(self, lcoords, *args, z=None, topo=None, **kwargs):
        topo = self.topology().to_numpy() if topo is None else topo
        gnum = topo_to_gnum(topo, self.NDOFN)
        lcoords = atleast2d(lcoords)
        dshp = self.shape_function_derivatives(lcoords)
        ecoords = self.local_coordinates(topo=topo)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)
        dofsol = self.pointdata.dofsol.to_numpy()
        dofsol = atleast3d(dofsol, back=True)
        return self.model_strains(dofsol, gnum, B)

    def stresses_at(self, *args, z=None, topo=None, **kwargs):
        """
        Returns stresses for every node of every element.
        """
        # The next code line returns either generalized strains
        # or material strains, depending on the value of z
        strns = self.strains_at(*args, z=z, topo=topo, **kwargs)
        C = self.elastic_material_stiffness_matrix()
        if "HMH" in args:
            return self.HMH(stresses_from_strains(C, strns))
        return stresses_from_strains(C, strns)

    def stresses_at_nodes(self, *args, **kwargs):
        """
        Returns stresses for every node of every element.
        """
        return self.stresses_at(self.lcoords(), *args, **kwargs)

    def stresses_at_centers(self, *args, **kwargs):
        """
        Returns stresses at the centre of every element.
        """
        return self.stresses_at(self.lcenter(), *args, **kwargs)



import numpy as np
from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools import getallfromkwargs

from polymesh.space import StandardFrame, PointCloud
from polymesh.grid import gridH8 as grid
from polymesh.utils.topology.tr import H8_to_L2
from polymesh.utils.space import index_of_closest_point, frames_of_lines

from sigmaepsilon import BeamSection
from sigmaepsilon import Structure, LineMesh, PointData
from sigmaepsilon.fem.cells import B2 as Beam


def console_grid_bernoulli() -> Structure:
    """
    Returns a simple console as a grid of 2-noded Bernoulli beams.

    Example
    -------
    >>> from sigmaepsilon.examples import console_grid_bernoulli
    >>> structure = console_grid_bernoulli()
    >>> structure.linsolve()
    """
    # units in kN and cm

    Lx, Ly, Lz = 100.0, 15.0, 10.0  # size of the grid
    nx, ny, nz = 4, 2, 2  # desity of the grid

    d = 1.2  # outer diameter of the tube
    t = 0.4  # thickness of the tube

    Ex = 21000.0
    nu = 0.25
    fy = 35.0

    loads = LinkedDeepDict()

    # load case 1
    loads["LC1", "position"] = [Lx, 0, 0]
    loads["LC1", "value"] = [0, 0, 2.0, 0, 0, 0]

    # load case 2
    loads["LC2", "position"] = [Lx, Ly / 2, 0]
    loads["LC2", "value"] = [0, 2.0, 0, 0, 0, 0]

    # load case 3
    loads["LC3", "position"] = [Lx, -Ly / 2, 0]
    loads["LC3", "value"] = [0, -2.0, 0, 0, 0, 0]

    gridparams = {
        "size": (Lx, Ly, Lz),
        "shape": (nx, ny, nz),
        "origo": (0, 0, 0),
        "start": 0,
    }

    coords, topo = grid(**gridparams)
    coords, topo = H8_to_L2(coords, topo)

    GlobalFrame = StandardFrame(dim=3)

    points = PointCloud(coords, frame=GlobalFrame).centralize()
    dx = -np.array([points[:, 0].min(), 0.0, 0.0])
    points.move(dx)
    coords = points.show()

    section = BeamSection("CHS", d=d, t=t, n=16)
    section.calculate_section_properties()
    section_props = section.section_properties
    A, Ix, Iy, Iz = getallfromkwargs(["A", "Ix", "Iy", "Iz"], **section_props)

    G = Ex / (2 * (1 + nu))
    Hooke = np.array(
        [[Ex * A, 0, 0, 0], [0, G * Ix, 0, 0], [0, 0, Ex * Iy, 0], [0, 0, 0, Ex * Iz]]
    )

    # essential boundary conditions
    ebcinds = np.where(coords[:, 0] < 1e-12)[0]
    fixity = np.zeros((coords.shape[0], 6)).astype(bool)
    fixity[ebcinds, :] = True

    # natural boundary conditions
    nLoadCase = len(loads)
    nodal_loads = np.zeros((coords.shape[0], 6, nLoadCase))
    for iLC, key in enumerate(loads):
        x = loads[key]["position"]
        f = loads[key]["value"]
        iN = index_of_closest_point(coords, np.array(x))
        loads[key]["node"] = iN
        nodal_loads[iN, :, iLC] = f

    # pointdata
    pd = PointData(coords=coords, frame=GlobalFrame, loads=nodal_loads, fixity=fixity)

    # celldata
    frames = frames_of_lines(coords, topo)
    cd = Beam(topo=topo, frames=frames)

    # set up mesh and structure
    mesh = LineMesh(pd, cd, model=Hooke, frame=GlobalFrame)
    structure = Structure(mesh=mesh)

    return structure



import numpy as np

from dewloosh.core.meta import ABCMeta_Weak

from .dofmap import DOF


class ABCMeta_FemMesh(ABCMeta_Weak):
    """
    Meta class for PointData and CellData classes.

    It merges attribute maps with those of the parent classes.

    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)
        if hasattr(cls, "dofs"):
            cls.NDOFN = len(cls.dofs)
            cls.dofmap = np.array(DOF.dofmap(cls.dofs), dtype=int)
        return cls


class ABC_FemMesh(metaclass=ABCMeta_FemMesh):
    """Helper class that provides a standard way to create an ABC using
    inheritance.
    """

    __slots__ = ()



from typing import Union

from scipy.sparse import coo_matrix as coo_np
from scipy.sparse import csr_matrix as csr_np
from scipy.sparse import csc_matrix as csc_np
from scipy.sparse import spmatrix
from scipy.sparse import isspmatrix as isspmatrix_np
import numpy as np
from numpy import ndarray
from numba import njit, prange
from numba.typed import Dict as nbDict
from numba import types as nbtypes
import awkward as ak

from neumann.linalg.sparse import JaggedArray

__cache = True

nbint64 = nbtypes.int64
nbuint64 = nbtypes.uint64


def index_mappers(*args, **kwargs) -> dict:
    if isinstance(args[0], (ndarray, JaggedArray)):
        return index_mappers_dense(*args, **kwargs)
    elif isspmatrix_np(args[0]):
        return index_mappers_sparse(*args, **kwargs)
    raise TypeError(f"Invalid type {type(args[0])} for the first positional argument!")


def index_mappers_dense(
    inds: Union[JaggedArray, ndarray],
    N: int = None,
    return_trivial: bool = False,
    return_inverse: bool = False,
) -> dict:
    imin, imax = np.min(inds), np.max(inds)
    loc_to_glob = array_to_dict(np.unique(inds), 0)
    N = imax if N is None else N
    glob_to_loc = invert_mapping(loc_to_glob) if return_inverse else None
    if imin == 0 and imax == (len(loc_to_glob) - 1) == (N - 1):
        if not return_trivial:
            loc_to_glob = None
            glob_to_loc = None
    if return_inverse:
        return loc_to_glob, glob_to_loc
    return loc_to_glob


def index_mappers_sparse(
    A: spmatrix, return_trivial: bool = False, return_inverse: bool = False
) -> dict:
    if isinstance(A, coo_np):
        inds = A.row
    elif isinstance(A, csr_np) or isinstance(A, csc_np):
        inds = A.indices
    else:
        raise NotImplementedError
    return index_mappers_dense(inds, A.shape[0], return_trivial, return_inverse)


def box_dof_numbering(
    gnum: Union[JaggedArray, ndarray], glob_to_loc: nbDict = None
) -> ndarray:
    if glob_to_loc is None:
        return gnum
    elif isinstance(glob_to_loc, nbDict):
        return box_indices_2d(gnum, glob_to_loc)
    raise NotImplementedError


def box_spmatrix(spmat: spmatrix, glob_to_loc: nbDict = None):
    if isinstance(spmat, coo_np):
        return box_coo(spmat, glob_to_loc)
    elif isinstance(spmat, csr_np) or isinstance(spmat, csc_np):
        return box_csr(spmat, glob_to_loc)
    else:
        raise NotImplementedError("Unknown type : {}".format(type(spmat)))


def box_coo(coo: coo_np, glob_to_loc: nbDict = None, sum_duplicates=True):
    if sum_duplicates:
        coo.sum_duplicates()
    if glob_to_loc is None:
        return coo
    elif isinstance(glob_to_loc, nbDict):
        N = len(glob_to_loc)
        coo.row = box_indices1d(coo.row, glob_to_loc)
        coo.col = box_indices1d(coo.col, glob_to_loc)
        coo._shape = (N, N)
        return coo
    elif isinstance(glob_to_loc, ndarray):
        """
        inds = np.where(glob_to_loc[coo.row] >= 0)[0]
        coo.row = coo.row[inds]
        coo.col = coo.col[inds]
        coo.data = coo.data[inds]
        pass
        """
    raise NotImplementedError


def box_csr(csr: csr_np, glob_to_loc: nbDict = None):
    """
    FIXME. This fails.
    """
    raise RuntimeWarning("FIXME. THIS FAILS!")
    if glob_to_loc is None:
        return csr
    elif isinstance(glob_to_loc, nbDict):
        N = len(glob_to_loc)
        csr.indices, csr.indptr = _box_csr_data(csr.indices, csr.indptr, glob_to_loc)
        csr._shape = (N, N)
        return csr
    raise NotImplementedError


@njit(nogil=True, parallel=True, cache=__cache)
def _box_csr_data(indices: np.ndarray, indptr: np.ndarray, glob_to_loc: nbDict):
    """
    FIXME. This fails.
    """
    raise RuntimeWarning("FIXME. THIS FAILS!")
    N = len(glob_to_loc)
    indices_ = np.zeros(indices.shape[0], dtype=indices.dtype)
    indptr_ = np.zeros(N + 1, dtype=indptr.dtype)
    widths_ = np.zeros(N, dtype=indptr.dtype)
    for i in prange(indices.shape[0]):
        indices_[i] = glob_to_loc[indices[i]]
    for i in prange(N):
        widths_[i] = indptr[i + 1] - indptr[i]
    indptr_[1:] = widths_
    return indices_, indptr_


def box_csc(csc: csc_np, glob_to_loc: nbDict = None):
    return box_csr(csc, glob_to_loc)


def box_rhs(f: np.ndarray, loc_to_glob: nbDict = None):
    if loc_to_glob is None:
        return f
    elif isinstance(loc_to_glob, nbDict):
        return _box(f, loc_to_glob)


def unbox_lhs(u: np.ndarray, loc_to_glob: nbDict = None, N: int = None):
    if loc_to_glob is None:
        return u
    elif isinstance(loc_to_glob, nbDict):
        if N is None:
            N = np.max(list(loc_to_glob.values())) + 1
        if len(u.shape) == 1:
            u = u.reshape(len(u), 1)
        return _unbox(u, loc_to_glob, N)


@njit(nogil=True, parallel=True, cache=__cache)
def _unbox(a: np.ndarray, loc_to_glob: nbDict, N: int) -> ndarray:
    res = np.zeros((N, a.shape[1]), dtype=a.dtype)
    for i in prange(len(loc_to_glob)):
        res[loc_to_glob[nbuint64(i)], :] = a[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _box(a: np.ndarray, loc_to_glob: nbDict) -> ndarray:
    N = len(loc_to_glob)
    res = np.zeros((N, a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :] = a[loc_to_glob[nbuint64(i)], :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def box_indices1d(indices: np.ndarray, glob_to_loc: nbDict) -> ndarray:
    """
    Returns an 1d index array, renumbered according to the
    mapping introduced by the mapping `glob_to_loc`.
    """
    res = np.zeros_like(indices)
    for i in prange(res.shape[0]):
        res[i] = glob_to_loc[nbuint64(indices[i])]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def box_indices_2d_regular(indices: np.ndarray, glob_to_loc: nbDict) -> ndarray:
    res = np.zeros_like(indices)
    for i in prange(res.shape[0]):
        for j in prange(res.shape[1]):
            res[i, j] = glob_to_loc[nbuint64(indices[i, j])]
    return res


@njit(nogil=True, parallel=False, cache=False)
def box_indices_2d_jagged(indices: JaggedArray, glob_to_loc: nbDict, builder):
    for i in range(len(indices)):
        inds = indices[i]
        builder.begin_list()
        for j in range(len(inds)):
            builder.integer(glob_to_loc[nbuint64(inds[j])])
        builder.end_list()


def box_indices_2d(
    indices: Union[ndarray, JaggedArray], glob_to_loc: nbDict
) -> Union[ndarray, JaggedArray]:
    if isinstance(indices, ndarray):
        return box_indices_2d_regular(indices, glob_to_loc)
    else:
        builder = ak.ArrayBuilder()
        box_indices_2d_jagged(indices.to_ak(), glob_to_loc, builder)
        return builder.snapshot()


@njit(nogil=True, cache=__cache)
def array_to_dict(a: np.ndarray, firstindex: int = 0):
    res = nbDict.empty(key_type=nbuint64, value_type=nbuint64)
    for i in range(len(a)):
        ui = nbuint64(firstindex + i)
        res[ui] = nbuint64(a[i])
    return res


@njit(nogil=True, cache=__cache)
def invert_mapping(imap: nbDict):
    res = nbDict.empty(key_type=nbuint64, value_type=nbuint64)
    for i in imap:
        ui = nbuint64(imap[i])
        res[ui] = nbuint64(i)
    return res



from polymesh.cells import H27 as HexaHedron
from polymesh.utils.cells.gauss import Gauss_Legendre_Hex_Grid

from ..material.solid3d import Solid3d
from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class H27(ABC, Solid3d, HexaHedron, FiniteElement):
    qrule = "full"
    quadrature = {
        "full": Gauss_Legendre_Hex_Grid(3, 3, 3),
        "selective": {(0, 1, 2): "full", (3, 4, 5): "reduced"},
        "reduced": Gauss_Legendre_Hex_Grid(2, 2, 2),
    }
